from pyglm.glm import vec3
from pyxpg import CullMode, PolygonMode

from ambra.config import CameraConfig, Config, GuiConfig, PlaybackConfig, RendererConfig
from ambra.geometry import (
    create_arrow,
    create_axis2d_lines_and_colors,
    create_axis3d_lines_and_colors,
    create_capped_cone,
    create_capped_cylinder,
    create_cone,
    create_cube,
    create_cube_edges,
    create_cylinder,
    create_disk,
    create_normal_lines,
    create_sphere,
)
from ambra.primitives3d import Lines, Mesh
from ambra.stages import two_directional_lights_and_uniform_environment_light
from ambra.viewer import Viewer

viewer = Viewer(
    config=Config(
        playback=PlaybackConfig(
            playing=True,
        ),
        camera=CameraConfig(
            position=vec3(3),
            target=vec3(0),
        ),
        gui=GuiConfig(
            playback=True,
            stats=True,
            inspector=True,
        ),
        renderer=RendererConfig(
            msaa_samples=4,
        ),
    ),
)

cull_mode = CullMode.BACK
polygon_mode = PolygonMode.FILL

cube_v, cube_n, cube_f = create_cube()
cube = Mesh(cube_v, cube_f, cube_n, cull_mode=cull_mode, polygon_mode=polygon_mode, translation=(-2.5, 0, 0))
cube_normals = Lines(*create_normal_lines(cube_v, cube_n, length=0.05), line_width=2, translation=(-2.5, 0, 0))

cone_v, cone_n, cone_f = create_cone()
cone = Mesh(cone_v, cone_f, cone_n, cull_mode=cull_mode, polygon_mode=polygon_mode, translation=(2.5, 0, 0))
cone_normals = Lines(*create_normal_lines(cone_v, cone_n, length=0.05), line_width=2, translation=(2.5, 0, 0))

capped_cone_v, capped_cone_n, capped_cone_f = create_capped_cone()
capped_cone = Mesh(
    capped_cone_v,
    capped_cone_f,
    capped_cone_n,
    cull_mode=cull_mode,
    polygon_mode=polygon_mode,
    translation=(2.5, -2.5, 0),
)
capped_cone_normals = Lines(
    *create_normal_lines(capped_cone_v, capped_cone_n, length=0.05), line_width=2, translation=(2.5, -2.5, 0)
)

disk_v, disk_n, disk_f = create_disk()
disk = Mesh(disk_v, disk_f, disk_n, cull_mode=cull_mode, polygon_mode=polygon_mode, translation=(0, 2.5, 0))
disk_normals = Lines(*create_normal_lines(disk_v, disk_n, length=0.05), line_width=2, translation=(0, 2.5, 0))

cylinder_v, cylinder_n, cylinder_f = create_cylinder()
cylinder = Mesh(
    cylinder_v, cylinder_f, cylinder_n, cull_mode=cull_mode, polygon_mode=polygon_mode, translation=(0, -2.5, 0)
)
cylinder_normals = Lines(
    *create_normal_lines(cylinder_v, cylinder_n, length=0.05), line_width=2, translation=(0, -2.5, 0)
)

cap_cylinder_v, cap_cylinder_n, cap_cylinder_f = create_capped_cylinder()
cap_cylinder = Mesh(
    cap_cylinder_v,
    cap_cylinder_f,
    cap_cylinder_n,
    cull_mode=cull_mode,
    polygon_mode=polygon_mode,
    translation=(-2.5, -2.5, 0),
)
cap_cylinder_normals = Lines(
    *create_normal_lines(cap_cylinder_v, cap_cylinder_n, length=0.05), line_width=2, translation=(-2.5, -2.5, 0)
)

sphere_v, sphere_n, sphere_f = create_sphere()
sphere = Mesh(sphere_v, sphere_f, sphere_n, cull_mode=cull_mode, polygon_mode=polygon_mode, translation=(-2.5, 2.5, 0))
sphere_normals = Lines(*create_normal_lines(sphere_v, sphere_n, length=0.05), line_width=2, translation=(-2.5, 2.5, 0))

arrow_v, arrow_n, arrow_f = create_arrow()
arrow = Mesh(arrow_v, arrow_f, arrow_n, cull_mode=cull_mode, polygon_mode=polygon_mode, translation=(2.5, 2.5, 0))
arrow_normals = Lines(*create_normal_lines(arrow_v, arrow_n, length=0.05), line_width=2, translation=(2.5, 2.5, 0))

ax2d = Lines(*create_axis2d_lines_and_colors(length=0.5), line_width=2, translation=(0, 0, 1.5))
ax3d = Lines(*create_axis3d_lines_and_colors(), line_width=2, translation=(0, 0, 0))
cube_edges = Lines(*create_cube_edges(), line_width=2, translation=(0, 0, 0))


viewer.scene.objects.extend(
    [
        cube,
        cube_normals,
        cone,
        cone_normals,
        capped_cone,
        capped_cone_normals,
        disk,
        disk_normals,
        cylinder,
        cylinder_normals,
        cap_cylinder,
        cap_cylinder_normals,
        sphere,
        sphere_normals,
        arrow,
        arrow_normals,
        ax2d,
        ax3d,
        cube_edges,
        *two_directional_lights_and_uniform_environment_light(),
    ]
)

viewer.run()



# Copyright Dario Mylonopoulos
# SPDX-License-Identifier: MIT

from queue import Empty, Queue
from time import perf_counter_ns
from typing import Any, Callable, List, Optional, Tuple

import numpy as np
from numpy.typing import NDArray
from pyglm.glm import dvec2, ivec2, normalize, vec3
from pyxpg import (
    Action,
    AllocType,
    BorderColor,
    BufferUsageFlags,
    DescriptorSetBinding,
    DescriptorType,
    Device,
    DeviceFeatures,
    Format,
    Gui,
    Image,
    ImageLayout,
    ImageUsageFlags,
    Instance,
    Key,
    LogLevel,
    MemoryHeapFlags,
    Modifiers,
    MouseButton,
    Sampler,
    SamplerAddressMode,
    Stage,
    SwapchainOutOfDateError,
    SwapchainStatus,
    Window,
    imgui,
    process_events,
    set_log_level,
)

from .config import Config
from .gpu_property import (
    GpuBufferView,
    GpuImageView,
    GpuPreuploadedArrayProperty,
    GpuPreuploadedProperty,
    GpuProperty,
    GpuStreamingProperty,
)
from .headless import HeadlessSwapchain, HeadlessSwapchainFrame
from .keybindings import KeyMap
from .lights import LIGHT_TYPES_INFO
from .renderer import FrameInputs, Renderer
from .scene import Object, Scene, Widget
from .server import Client, Message, RawMessage, Server, parse_builtin_messages
from .utils.descriptors import (
    create_descriptor_layout_pool_and_sets,
    create_descriptor_pool_and_sets,
)
from .utils.gpu import UploadableBuffer
from .utils.lru_pool import LRUPool
from .utils.ring_buffer import RingBuffer
from .viewport import Playback, Rect, Viewport


class Viewer:
    def __init__(
        self,
        title: str = "ambra",
        config: Optional[Config] = None,
        key_map: Optional[KeyMap] = None,
    ):
        # Config
        config = config if config is not None else Config()

        # Key bindings
        self.key_map = key_map if key_map is not None else KeyMap()

        # Logging
        if config.log_level is not None:
            set_log_level(LogLevel(config.log_level.value))

        # Viewer
        self.running = False

        # Device
        self.instance = Instance(
            version=(1, 1),
            presentation=config.window,
            enable_validation_layer=config.enable_validation_layer,
            enable_synchronization_validation=config.enable_synchronization_validation,
            enable_gpu_based_validation=config.enable_gpu_based_validation,
            enable_shader_debug_printf=config.enable_shader_debug_printf,
            shader_debug_printf_buffer_size=config.shader_debug_printf_buffer_size,
        )

        self.device = Device(
            self.instance,
            version=(1, 1),
            required_features=DeviceFeatures.SYNCHRONIZATION_2
            | DeviceFeatures.DYNAMIC_RENDERING
            | DeviceFeatures.DESCRIPTOR_INDEXING
            | DeviceFeatures.STORAGE_IMAGE_READ_WRITE_WITHOUT_FORMAT,
            optional_features=DeviceFeatures.RAY_QUERY
            | DeviceFeatures.HOST_QUERY_RESET
            | DeviceFeatures.WIDE_LINES
            | DeviceFeatures.TIMELINE_SEMAPHORES
            | DeviceFeatures.SHADER_DRAW_PARAMETERS
            | DeviceFeatures.SHADER_FLOAT16_INT8
            | DeviceFeatures.SHADER_INT16
            | DeviceFeatures.SHADER_INT64
            | DeviceFeatures.SHADER_SUBGROUP_EXTENDED_TYPES
            | DeviceFeatures.STORAGE_8BIT
            | DeviceFeatures.STORAGE_16BIT
            | DeviceFeatures.DRAW_INDIRECT_COUNT
            | DeviceFeatures.MESH_SHADER
            | DeviceFeatures.FRAGMENT_SHADER_BARYCENTRIC
            | DeviceFeatures.SUBGROUP_SIZE_CONTROL,
            presentation=config.window,
            force_physical_device_index=0xFFFFFFFF
            if config.force_physical_device_index is None
            else config.force_physical_device_index,
            prefer_discrete_gpu=config.prefer_discrete_gpu,
        )

        # Window
        self.window: Optional[Window] = None
        if config.window:
            self.window = Window(
                self.device,
                title,
                config.window_width,
                config.window_height,
                x=config.window_x,
                y=config.window_y,
                preferred_frames_in_flight=config.preferred_frames_in_flight,
                preferred_swapchain_usage_flags=ImageUsageFlags.COLOR_ATTACHMENT
                | ImageUsageFlags.TRANSFER_DST
                | ImageUsageFlags.TRANSFER_SRC
                | ImageUsageFlags.STORAGE,
                vsync=config.vsync,
            )
            self.window.set_callbacks(
                draw=self.on_draw,
                key_event=self.on_key,
                mouse_button_event=lambda pos, button, action, modifiers: self.on_mouse_button(
                    ivec2(pos), button, action, modifiers
                ),
                mouse_move_event=lambda pos: self.on_mouse_move(ivec2(pos)),
                mouse_scroll_event=lambda pos, scroll: self.on_scroll(ivec2(pos), dvec2(scroll)),
            )

        # Headless swapchain for screenshots and videos
        output_format = Format.R8G8B8A8_UNORM if self.window is None else self.window.swapchain_format
        self.headless_swapchain = HeadlessSwapchain(self.device, 2, output_format)

        if self.window is not None:
            render_width, render_height = self.window.fb_width, self.window.fb_height
            num_frames_in_flight = self.window.num_frames

            # GUI
            self.gui = Gui(self.window, config.gui.default_font_size, config.gui.default_font_preference)
        else:
            render_width, render_height = config.window_width, config.window_height
            num_frames_in_flight = self.headless_swapchain.num_frames_in_flight

            # GUI
            self.gui = Gui(
                self.instance,
                self.device,
                num_frames_in_flight,
                output_format,
                config.gui.default_font_size,
                config.gui.default_font_preference,
            )

        self.multiviewport = config.gui.multiviewport
        self.gui.set_ini_filename(config.gui.ini_filename)

        self.gui_show_stats = config.gui.stats
        self.gui_show_inspector = config.gui.inspector
        self.gui_show_playback = config.gui.playback
        self.gui_show_renderer = config.gui.renderer
        self.gui_selected_obj: Optional[Object] = None
        self.gui_selected_gpu_property: Optional[GpuProperty[Any]] = None
        self.gui_playback_slider_held = False

        # Disable ImGui asserts
        imgui.get_io().config_error_recovery_enable_assert = False

        # Make ImGui automatically scale fonts on DPI change events
        imgui.get_io().config_dpi_scale_fonts = True

        style = imgui.get_style()
        style.font_scale_main = config.gui.font_scale

        # Renderer
        self.renderer = Renderer(
            self.device,
            render_width,
            render_height,
            num_frames_in_flight,
            output_format,
            self.multiviewport,
            config.gui.max_viewport_count,
            config.renderer,
        )

        # Playback
        self.playback = Playback(config.playback)
        self.last_frame_timestamp = 0

        # Stats
        self.frame_time_index = 0
        self.frame_times = np.zeros(config.stats_frame_time_count, np.float32)

        # Scene
        self.scene = Scene("scene")

        # Widgets
        self.widgets: List[Widget] = []

        # Viewport
        if self.multiviewport:
            if config.gui.initial_number_of_viewports > config.gui.max_viewport_count:
                raise ValueError(
                    "config.gui.initial_number_of_viewports must be less than or equal to config.gui.max_viewport_count"
                )
            self.viewport_sampler = Sampler(
                self.device,
                u=SamplerAddressMode.CLAMP_TO_BORDER,
                v=SamplerAddressMode.CLAMP_TO_BORDER,
                border_color=BorderColor.FLOAT_OPAQUE_BLACK,
            )
            self.viewport_descriptor_layout, self.viewport_descriptor_pool, self.viewport_descriptor_sets = (
                create_descriptor_layout_pool_and_sets(
                    self.device,
                    [
                        DescriptorSetBinding(1, DescriptorType.COMBINED_IMAGE_SAMPLER, stage_flags=Stage.FRAGMENT),
                    ],
                    config.gui.max_viewport_count,
                )
            )
            num_viewports = config.gui.initial_number_of_viewports
            max_viewports = config.gui.max_viewport_count
        else:
            num_viewports = 1
            max_viewports = 1

        self.viewport_scene_descriptor_pool, self.viewport_scene_descriptor_sets = create_descriptor_pool_and_sets(
            self.device,
            self.renderer.scene_descriptor_set_layout,
            max_viewports * self.renderer.num_frames_in_flight,
            name="viewport-scene-descriptor-sets",
        )
        self.viewports: List[Viewport] = []
        for viewport_index in range(min(num_viewports, max_viewports)):
            scene_descriptor_sets = RingBuffer(
                self.viewport_scene_descriptor_sets[
                    viewport_index * self.renderer.num_frames_in_flight : (viewport_index + 1)
                    * self.renderer.num_frames_in_flight
                ]
            )
            scene_uniform_buffers = RingBuffer(
                [
                    UploadableBuffer(
                        self.device, self.renderer.scene_constants_dtype.itemsize, BufferUsageFlags.UNIFORM
                    )
                    for _ in range(self.renderer.num_frames_in_flight)
                ]
            )
            scene_light_buffers = RingBuffer(
                [
                    [
                        UploadableBuffer(
                            self.device, info.size * self.renderer.max_lights_per_type, BufferUsageFlags.STORAGE
                        )
                        for info in LIGHT_TYPES_INFO
                    ]
                    for _ in range(self.renderer.num_frames_in_flight)
                ]
            )

            for s, buf, light_bufs in zip(scene_descriptor_sets, scene_uniform_buffers, scene_light_buffers):
                s.write_buffer(buf, DescriptorType.UNIFORM_BUFFER, 0)
                s.write_sampler(self.renderer.shadow_sampler, 1)
                s.write_sampler(self.renderer.linear_sampler, 2)
                s.write_image(
                    self.renderer.zero_cubemap, ImageLayout.SHADER_READ_ONLY_OPTIMAL, DescriptorType.SAMPLED_IMAGE, 3
                )
                s.write_image(
                    self.renderer.zero_cubemap, ImageLayout.SHADER_READ_ONLY_OPTIMAL, DescriptorType.SAMPLED_IMAGE, 4
                )
                s.write_image(
                    self.renderer.ggx_lut, ImageLayout.SHADER_READ_ONLY_OPTIMAL, DescriptorType.SAMPLED_IMAGE, 5
                )

                for i, light_buf in enumerate(light_bufs):
                    s.write_buffer(light_buf, DescriptorType.STORAGE_BUFFER, 6, i)
                for i in range(self.renderer.max_shadow_maps):
                    s.write_image(
                        self.renderer.zero_image,
                        ImageLayout.SHADER_READ_ONLY_OPTIMAL,
                        DescriptorType.SAMPLED_IMAGE,
                        6 + len(light_bufs),
                        i,
                    )

            scene_constants = np.zeros((1,), self.renderer.scene_constants_dtype)

            img = None
            texture = None
            if self.multiviewport:
                usage_flags = ImageUsageFlags.SAMPLED | ImageUsageFlags.COLOR_ATTACHMENT
                if self.renderer.supports_path_tracing:
                    usage_flags |= ImageUsageFlags.STORAGE
                img = Image(
                    self.device,
                    render_width,
                    render_height,
                    output_format,
                    usage_flags,
                    AllocType.DEVICE,
                    name=f"viewport-{viewport_index}",
                )

                s = self.viewport_descriptor_sets[viewport_index]
                s.write_combined_image_sampler(img, ImageLayout.SHADER_READ_ONLY_OPTIMAL, self.viewport_sampler, 0)
                texture = imgui.Texture(s)

            self.viewports.append(
                Viewport(
                    rect=Rect(0, 0, render_width, render_height),
                    playback=self.playback,
                    camera_config=config.camera,
                    handedness=config.handedness,
                    world_up=normalize(vec3(config.world_up)),
                    scene_descriptor_sets=scene_descriptor_sets,
                    scene_uniform_buffers=scene_uniform_buffers,
                    scene_light_buffers=scene_light_buffers,
                    scene_constants=scene_constants,
                    image=img,
                    imgui_texture=texture,
                    name=f"viewport-{viewport_index}",
                )
            )
        self.active_viewport = self.viewports[0] if self.viewports else None

        # Server
        self.server = Server(self.on_raw_message_async, config.server)
        self.server_message_queue: Queue[Tuple[Client, RawMessage]] = Queue()

        # Config
        self.wait_events = config.wait_events

    def on_key(self, key: Key, action: Action, modifiers: Modifiers) -> None:
        if imgui.get_io().want_capture_keyboard:
            return

        # Press
        if action == Action.PRESS:
            if self.key_map.toggle_play_pause.is_active(key, modifiers):
                self.playback.toggle_play_pause()
                # If just restarted playing, reset last_frame_timestamp
                if self.playback.playing:
                    self.last_frame_timestamp = perf_counter_ns()
            if self.key_map.exit.is_active(key, modifiers):
                self.running = False
            if self.key_map.toggle_path_tracer.is_active(key, modifiers):
                self.renderer.toggle_path_tracer()

        # Press + repeat
        if action == Action.PRESS or action == Action.REPEAT:
            if self.key_map.next_frame.is_active(key, modifiers):
                self.playback.set_frame(self.playback.current_frame + 1)
            if self.key_map.previous_frame.is_active(key, modifiers):
                self.playback.set_frame(self.playback.current_frame - 1)

    def on_mouse_button(
        self,
        position: ivec2,
        button: MouseButton,
        action: Action,
        modifiers: Modifiers,
    ) -> None:
        if imgui.get_io().want_capture_mouse:
            return

        if self.active_viewport is not None:
            if action == Action.PRESS:
                if self.key_map.camera_rotate.is_active(button, modifiers):
                    self.active_viewport.on_rotate_press(position)
                if self.key_map.camera_pan.is_active(button, modifiers):
                    self.active_viewport.on_pan_press(position)
            if action == Action.RELEASE:
                if self.key_map.camera_rotate.button == button:
                    self.active_viewport.on_rotate_release()
                if self.key_map.camera_pan.button == button:
                    self.active_viewport.on_pan_release()

    def on_mouse_move(self, position: ivec2) -> None:
        if self.active_viewport is not None:
            self.active_viewport.on_move(position)

    def on_scroll(self, position: ivec2, scroll: dvec2) -> None:
        if imgui.get_io().want_capture_mouse:
            return

        if self.active_viewport is not None:
            # Callbacks are only called if a window is present
            assert self.window is not None
            modifiers = self.window.get_modifiers_state()

            if modifiers == self.key_map.camera_zoom_modifiers:
                self.active_viewport.on_zoom(scroll)
            if modifiers == self.key_map.camera_zoom_move_modifiers:
                self.active_viewport.on_zoom_with_movement(scroll)

    def on_resize(self, width: int, height: int) -> None:
        pass

    def _render(self, render_to_window: bool) -> None:
        # Begin frame
        if render_to_window:
            assert self.window is not None
            try:
                frame = self.window.begin_frame()
            except SwapchainOutOfDateError:
                return
            frame_inputs = FrameInputs(frame.image, frame.command_buffer, frame.transfer_command_buffer, [], [])
        else:
            frame_inputs = self.headless_swapchain.begin_frame()

        # Begin recording on command buffers
        frame_inputs.command_buffer.begin()
        if frame_inputs.transfer_command_buffer is not None:
            frame_inputs.transfer_command_buffer.begin()

        # Draw GUI
        with self.gui.frame():
            # Pre-render gui with things that can affect the rendering of this frame.
            # Especially important when running with wait_events set to True because
            # otherwise we may not see the effect of camera movement and other scene
            # edits due to events that cause a single frame of rendering
            self.on_gui()

            # Render scene
            self.renderer.render(self.scene, self.viewports, frame_inputs)

            # Other parts of the gui have to happen after rendering for multiple reasons:
            # - Widgets can rely on properties being already loaded and uploaded
            #   which happens in render, they also will be created here on first use.
            # - Renderer UI for this frame will be up to date with latest object and property state.
            #
            # TODO: User defined widgets will still have a frame of lag. Also some of the imgui interactions
            # require multiple frames after an event to take effect. In practice to fully support wait_events
            # we anyways need to introduce an "animation_playing" or "num_frames_to_render" state to render
            # more than one frame per event.
            self.on_gui_after_render()

        # HACK: If not rendering to the window, we draw the GUI twice so that
        # imgui renders all windows, including those for which the size of the contents hasn't
        # been computed yet.
        if not render_to_window:
            with self.gui.frame():
                self.on_gui()
                self.on_gui_after_render()

        # Render GUI
        self.renderer.render_gui(frame_inputs, self.gui)

        if not render_to_window:
            self.headless_swapchain.get_current().issue_readback()

        # Submit transfer queue commands, if submitted
        if frame_inputs.transfer_command_buffer is not None:
            frame_inputs.transfer_command_buffer.end()

            # If there is no semaphore to signal, we assume there is no commands to submit.
            # Even if someone recorded commands, this would be a bug because
            # there is no way to know when those commands would complete.
            if frame_inputs.transfer_semaphores:
                self.device.transfer_queue.submit(
                    frame_inputs.transfer_command_buffer,
                    wait_semaphores=[(s.sem, s.wait_value, s.wait_stage) for s in frame_inputs.transfer_semaphores],
                    signal_semaphores=[
                        (s.sem, s.signal_value, s.signal_stage) for s in frame_inputs.transfer_semaphores
                    ],
                )

        # Close recording on the graphics commands
        frame_inputs.command_buffer.end()

        # End frame
        if render_to_window:
            assert self.window is not None
            self.window.end_frame(
                frame,
                additional_wait_semaphores=[
                    (s.sem, s.wait_value, s.wait_stage) for s in frame_inputs.additional_semaphores
                ],
                additional_signal_semaphores=[
                    (s.sem, s.signal_value, s.signal_stage) for s in frame_inputs.additional_semaphores
                ],
            )
        else:
            self.headless_swapchain.end_frame(frame_inputs)

        # Prefetch next frames
        self.renderer.prefetch()

    def _get_framebuffer_size(self) -> Tuple[int, int]:
        if self.window is not None:
            width = self.window.fb_width
            height = self.window.fb_height
        else:
            # In headless mode there is always a single viewport
            width = self.viewports[0].rect.width
            height = self.viewports[0].rect.height
        return width, height

    def render_image(self) -> NDArray[np.uint8]:
        # Set max time if not set
        if self.playback.max_time is None:
            self.playback.set_max_time(self.scene.end_animation_time(self.playback.frames_per_second))

        # Update scene
        self.scene.update(self.playback.current_time, self.playback.current_frame)

        # Get framebuffer size
        width, height = self._get_framebuffer_size()

        # Resize if needed
        self.headless_swapchain.ensure_size(width, height)

        # TODO: I think in the headless case we also need to resize the renderer
        # here, for things like depth buffer and msaa targets. Double check this
        # by making a test that renders in headless mode at different resolutions
        # on the same viewer.

        io = imgui.get_io()
        io.display_size = imgui.Vec2(width, height)
        io.delta_time = 1.0 / 60.0

        # Render to headless swapchain
        self._render(False)

        # Get current headless swapchain frame
        frame = self.headless_swapchain.get_current_and_advance()

        # Readback frame
        return frame.realize_readback()

    def render_video(self, on_frame: Callable[[NDArray[np.uint8]], bool]) -> None:
        # Set max time if not set
        if self.playback.max_time is None:
            self.playback.set_max_time(self.scene.end_animation_time(self.playback.frames_per_second))

        begin_frame_index = 0
        end_frame_index = begin_frame_index + self.playback.num_frames

        # Get framebuffer size
        width, height = self._get_framebuffer_size()

        # Resize if needed
        self.headless_swapchain.ensure_size(width, height)

        io = imgui.get_io()
        io.display_size = imgui.Vec2(width, height)
        io.delta_time = 1.0 / 60.0

        # Readback queue for pipelining rendering and readback.
        readback_queue: Queue[HeadlessSwapchainFrame] = Queue()

        def deque_frame() -> None:
            old_frame = readback_queue.get()
            img = old_frame.realize_readback()
            on_frame(img)

        for frame_index in range(begin_frame_index, end_frame_index):
            if readback_queue.qsize() >= self.headless_swapchain.num_frames_in_flight:
                deque_frame()

            self.playback.set_frame(frame_index)

            # Update scene
            self.scene.update(self.playback.current_time, self.playback.current_frame)

            # Render to headless swapchain
            self._render(False)

            # Get current headless swapchain frame
            frame = self.headless_swapchain.get_current_and_advance()

            # Enqueue for future readback
            readback_queue.put(frame)

        # Drain readback queue
        while not readback_queue.empty():
            deque_frame()

    def on_draw(self) -> None:
        # Callbacks are only called if a window is present
        assert self.window is not None

        # Compute dt
        timestamp = perf_counter_ns()
        dt = (timestamp - self.last_frame_timestamp) * 1e-9
        self.last_frame_timestamp = timestamp

        self.frame_times[self.frame_time_index] = dt
        self.frame_time_index = (self.frame_time_index + 1) % self.frame_times.size

        # Set max time if not set
        if self.playback.max_time is None:
            self.playback.set_max_time(self.scene.end_animation_time(self.playback.frames_per_second))

        # Step dt
        if self.playback.playing and not self.gui_playback_slider_held:
            self.playback.step(dt)

        # Update scene
        self.scene.update(self.playback.current_time, self.playback.current_frame)

        # Resize
        swapchain_status = self.window.update_swapchain()
        if swapchain_status == SwapchainStatus.MINIMIZED:
            return
        if swapchain_status == SwapchainStatus.RESIZED:
            width, height = self.window.fb_width, self.window.fb_height

            self.renderer.resize(width, height)

            if not self.multiviewport:
                # In the multi-viewport case rect resizing is handled as part of
                # the GUI update.
                self.viewports[0].resize(width, height)
            else:
                for viewport_index, viewport in enumerate(self.viewports):
                    assert viewport.image is not None
                    viewport.image.destroy()

                    usage_flags = ImageUsageFlags.SAMPLED | ImageUsageFlags.COLOR_ATTACHMENT
                    if self.renderer.supports_path_tracing:
                        usage_flags |= ImageUsageFlags.STORAGE
                    img = Image(
                        self.device,
                        width,
                        height,
                        self.renderer.output_format,
                        usage_flags,
                        AllocType.DEVICE,
                        name=f"viewport-{viewport_index}",
                    )

                    s = self.viewport_descriptor_sets[viewport_index]
                    s.write_combined_image_sampler(img, ImageLayout.SHADER_READ_ONLY_OPTIMAL, self.viewport_sampler, 0)
                    viewport.image = img

            self.on_resize(width, height)

        # Render to window
        self._render(True)

    def on_raw_message_async(self, client: Client, raw_message: RawMessage) -> None:
        self.server_message_queue.put((client, raw_message))
        if self.window is not None:
            self.window.post_empty_event()

    def on_raw_message(self, client: Client, raw_message: RawMessage) -> None:
        message = parse_builtin_messages(raw_message)
        if message is not None:
            self.on_message(client, message)

    def on_message(self, client: Client, message: Message) -> None:
        pass

    def gui_stats(self) -> None:
        imgui.set_next_window_pos(imgui.Vec2(10, 10))
        imgui.set_next_window_bg_alpha(0.3)
        if imgui.begin(
            "Stats",
            flags=imgui.WindowFlags.NO_TITLE_BAR
            | imgui.WindowFlags.ALWAYS_AUTO_RESIZE
            | imgui.WindowFlags.NO_RESIZE
            | imgui.WindowFlags.NO_SAVED_SETTINGS
            | imgui.WindowFlags.NO_MOVE
            | imgui.WindowFlags.NO_FOCUS_ON_APPEARING
            | imgui.WindowFlags.NO_NAV
            | imgui.WindowFlags.NO_MOUSE_INPUTS
            | imgui.WindowFlags.NO_SCROLLBAR,
        )[0]:
            avg_dt = self.frame_times.mean()
            avg_fps = 1.0 / avg_dt if avg_dt > 0 else 0.0
            last_dt = self.frame_times[self.frame_time_index]
            last_fps = 1.0 / last_dt if last_dt > 0 else 0.0
            imgui.text(f"{self.device.device_properties.device_name}")
            if self.window is not None:
                imgui.text(f"Window size:     [{self.window.fb_width}x{self.window.fb_height}]")
            imgui.text(f"FPS:             {avg_fps:6.2f} ({last_fps:6.2f})")
            imgui.text(f"Frame time (ms): {avg_dt * 1000.0:6.2f} ({last_dt * 1000.0:6.2f})")
            for i, (heap, stats) in enumerate(
                zip(self.device.memory_properties.memory_heaps, self.device.heap_statistics)
            ):
                if heap.flags & MemoryHeapFlags.VK_MEMORY_HEAP_DEVICE_LOCAL:
                    kind = "GPU"
                else:
                    kind = "CPU"
                imgui.text(
                    f"Heap {i} - {kind}: {stats.block_bytes / (1024 * 1024 * 1024):4.02f} / {heap.size / (1024 * 1024 * 1024):4.02f} GB"
                )
        imgui.end()

    def gui_playback(self) -> None:
        self.gui_playback_slider_held = False
        if imgui.begin("Playback")[0]:
            u, self.playback.playing = imgui.checkbox("Playing", self.playback.playing)
            # If just restarted playing, reset last_frame_timestamp
            if u and self.playback.playing:
                self.last_frame_timestamp = perf_counter_ns()
            _, self.playback.playback_speed_multiplier = imgui.drag_float(
                "Playback speed", self.playback.playback_speed_multiplier, 0.01, 0.01, 10.0, format="%.2fx"
            )
            _, frame = imgui.slider_int(
                "Frame",
                self.playback.current_frame,
                0,
                self.playback.num_frames - 1,
            )
            if imgui.is_item_active():
                self.gui_playback_slider_held = True
                self.playback.set_frame(frame)
            imgui.text(f"Time (s): {self.playback.current_time:7.3f} / {self.playback.max_time: 7.3f}")
            imgui.text(f"Playback FPS: {self.playback.frames_per_second:7.3f}")
        imgui.end()

    def gui_inspector(self) -> None:
        if imgui.begin("Inspector")[0]:
            for i, v in enumerate(self.viewports):
                if imgui.tree_node_ex(f"Camera - {v.name}##{i}" if self.multiviewport else "Camera"):
                    imgui.begin_disabled()
                    imgui.drag_float3(f"Position##{i}", tuple(v.camera.position()))
                    right, up, front = v.camera.right_up_front()
                    imgui.drag_float3(f"Right##{i}", tuple(right))
                    imgui.drag_float3(f"Up##{i}", tuple(up))
                    imgui.drag_float3(f"Front##{i}", tuple(front))
                    imgui.end_disabled()
                    imgui.tree_pop()
            imgui.separator()

            def pre(o: Object) -> bool:
                flags = imgui.TreeNodeFlags.OPEN_ON_ARROW | imgui.TreeNodeFlags.FRAME_PADDING
                if o.gui_expanded:
                    flags |= imgui.TreeNodeFlags.DEFAULT_OPEN
                if self.gui_selected_obj == o:
                    flags |= imgui.TreeNodeFlags.SELECTED
                if not o.children:
                    flags |= imgui.TreeNodeFlags.LEAF
                    flags |= imgui.TreeNodeFlags.BULLET

                imgui.push_style_var_im_vec2(imgui.StyleVar.FRAME_PADDING, imgui.Vec2(0, 3))
                o.gui_expanded = imgui.tree_node_ex(f"{o.name}##tree_node_{o.uid}", flags)
                imgui.pop_style_var()

                if imgui.is_item_clicked():
                    if self.gui_selected_obj == o:
                        self.gui_selected_obj = None
                    else:
                        self.gui_selected_obj = o

                imgui.same_line(imgui.get_content_region_max().x - 25)
                _, o.gui_enabled = imgui.checkbox(f"##enabled_{o.uid}", o.gui_enabled)

                return o.gui_expanded

            def post(o: Object) -> None:
                imgui.tree_pop()

            self.scene.visit_objects_pre_post(pre, post)

            imgui.separator()

            if self.gui_selected_obj is not None:
                self.gui_selected_obj.gui()

        imgui.end()

    def gui_renderer(self) -> None:
        if imgui.begin("Renderer")[0]:
            imgui.text("GPU properties:")
            imgui.indent(5)
            for i, p in enumerate(self.renderer.enabled_gpu_properties):
                s = imgui.selectable(f"{p.name}##{i}", self.gui_selected_gpu_property == p)
                if s:
                    if self.gui_selected_gpu_property == p:
                        self.gui_selected_gpu_property = None
                    else:
                        self.gui_selected_gpu_property = p
            imgui.indent(-5)

            imgui.separator()
            if self.gui_selected_gpu_property is not None:
                imgui.separator_text(type(self.gui_selected_gpu_property).__name__)
                if isinstance(self.gui_selected_gpu_property, GpuPreuploadedArrayProperty):
                    imgui.text(f"Frame size: {self.gui_selected_gpu_property.frame_size}")
                    imgui.text(f"Upload method: {self.gui_selected_gpu_property.upload_method}")
                    inv = ", ".join([str(i) for i in self.gui_selected_gpu_property.invalid_frames])
                    imgui.text(f"Invalid frames: {{{inv}}}")
                    imgui.text("Resource:")
                    imgui.indent()
                    imgui.text(f"{self.gui_selected_gpu_property.resource}")
                    imgui.text(f"{self.gui_selected_gpu_property.resource.alloc}")
                    imgui.unindent()
                if isinstance(self.gui_selected_gpu_property, GpuPreuploadedProperty):
                    imgui.text(f"Frame size: {self.gui_selected_gpu_property.max_frame_size}")
                    imgui.text(f"Batched: {self.gui_selected_gpu_property.batched}")
                    imgui.text(f"Async load: {self.gui_selected_gpu_property.async_load}")
                    imgui.text(f"Upload method: {self.gui_selected_gpu_property.upload_method}")
                    inv = ", ".join([str(i) for i in self.gui_selected_gpu_property.invalid_frames])
                    imgui.text(f"Invalid frames: {{{inv}}}")
                    imgui.text(f"Resources: {len(self.gui_selected_gpu_property.resources)}")
                    if isinstance(self.gui_selected_gpu_property.current, GpuBufferView):
                        imgui.indent()
                        imgui.text(f"{self.gui_selected_gpu_property.current.buffer}")
                        imgui.unindent()
                    elif isinstance(self.gui_selected_gpu_property.current, GpuImageView):
                        imgui.indent()
                        imgui.text(f"{self.gui_selected_gpu_property.current.image}")
                        imgui.unindent()
                elif isinstance(self.gui_selected_gpu_property, GpuStreamingProperty):
                    imgui.text(f"Upload method: {self.gui_selected_gpu_property.upload_method}")
                    imgui.text("CPU Resource:")
                    imgui.indent()
                    imgui.text(f"{self.gui_selected_gpu_property.cpu_buffers[0].buf}")
                    imgui.text(f"{self.gui_selected_gpu_property.cpu_buffers[0].buf.alloc}")
                    imgui.unindent()
                    imgui.text("GPU Resource:")
                    if self.gui_selected_gpu_property.gpu_resources:
                        imgui.indent()
                        v = self.gui_selected_gpu_property.gpu_resources[0].resource
                        imgui.text(f"{v}")
                        if isinstance(v, GpuBufferView):
                            imgui.text(f"{v.buffer.alloc}")
                        elif isinstance(v, GpuImageView):
                            imgui.text(f"{v.image.alloc}")
                        imgui.unindent()

                    def drawpool(name: str, pool: Optional[LRUPool[int, Any]], count: int) -> None:
                        if pool is None:
                            return
                        imgui.separator_text(name)
                        imgui.text("Map")
                        imgui.indent()
                        for lu_k, lu_v in pool.lookup.items():
                            imgui.text(f"{lu_k} {lu_v}")
                        imgui.unindent()

                        imgui.text("LRU")
                        imgui.indent()
                        i = 0
                        for lru_k, lru_v in pool.lru.items():
                            imgui.text(f"{lru_k} {lru_v}")
                            i += 1
                        for _ in range(i, count):
                            imgui.text("<EMPTY>")
                        imgui.unindent()

                        imgui.text("In Flight")
                        imgui.indent()
                        for if_v in pool.in_flight:
                            imgui.text(f"{if_v}")
                        imgui.unindent()

                        imgui.text("Prefetching")
                        imgui.indent()
                        i = 0
                        for pre_k, pre_v in pool.prefetch_store.items():
                            imgui.text(f"{pre_k} {pre_v}")
                            i += 1
                        for _ in range(i, pool.max_prefetch):
                            imgui.text("<EMPTY>")
                        imgui.unindent()

                        imgui.text("Generation indices")
                        imgui.indent()
                        for gen_k, gen_index in pool.current_generation.items():
                            imgui.text(f"{gen_k}: {gen_index}")
                        imgui.unindent()

                    drawpool(
                        "CPU",
                        self.gui_selected_gpu_property.cpu_pool,
                        len(self.gui_selected_gpu_property.cpu_buffers),
                    )
                    drawpool(
                        "GPU",
                        self.gui_selected_gpu_property.gpu_pool,
                        len(self.gui_selected_gpu_property.gpu_resources),
                    )

                    if self.gui_selected_gpu_property.cpu_pool or self.gui_selected_gpu_property.gpu_pool:
                        imgui.separator_text("Frame states")
                        imgui.spacing()

                        start = imgui.get_cursor_screen_pos()
                        dl = imgui.get_window_draw_list()

                        num_frames = self.gui_selected_gpu_property.property.num_frames
                        current_frame = self.gui_selected_gpu_property.property.current_frame_index

                        p_min = np.empty((num_frames, 2), np.float32)
                        p_max = np.empty((num_frames, 2), np.float32)
                        delta_x = 5 * np.arange(num_frames, dtype=np.float32)
                        p_min[:, 0] = start.x + delta_x
                        p_min[:, 1] = start.y
                        p_max[:, 0] = (start.x + 6) + delta_x
                        p_max[:, 1] = start.y + 20
                        dl.add_rect_batch(
                            p_min,
                            p_max,
                            np.array((0xFFFFFFFF,), np.uint32),
                            np.array((0.0,), np.float32),
                            np.array((1.0,), np.float32),
                        )

                        for (
                            (c_k, _),
                            c_v,
                        ) in self.gui_selected_gpu_property.cpu_pool.lookup.items():
                            cursor = imgui.Vec2(start.x + 5 * c_k, start.y)
                            color = 0xFF00FF00 if c_k >= current_frame else 0xFF0000FF
                            color = color if not c_v.prefetching else 0xFF00FFFF
                            dl.add_rect_filled(
                                imgui.Vec2(cursor.x + 1, cursor.y + 1),
                                imgui.Vec2(cursor.x + 5, cursor.y + 19),
                                color,
                            )
                        imgui.spacing()
                        imgui.spacing()
                        imgui.spacing()
                        imgui.spacing()
                        imgui.spacing()

                        if self.gui_selected_gpu_property.gpu_pool:
                            p_min[:, 1] = start.y + 22
                            p_max[:, 1] = start.y + 42
                            dl.add_rect_batch(
                                p_min,
                                p_max,
                                np.array((0xFFFFFFFF,), np.uint32),
                                np.array((0.0,), np.float32),
                                np.array((1.0,), np.float32),
                            )

                            for (
                                (g_k, _),
                                g_v,
                            ) in self.gui_selected_gpu_property.gpu_pool.lookup.items():
                                cursor = imgui.Vec2(start.x + 5 * g_k, start.y)
                                color = 0xFF00FF00 if g_k >= current_frame else 0xFF0000FF
                                color = color if not g_v.prefetching else 0xFF00FFFF
                                dl.add_rect_filled(
                                    imgui.Vec2(cursor.x + 1, cursor.y + 23),
                                    imgui.Vec2(cursor.x + 5, cursor.y + 41),
                                    color,
                                )
                            imgui.spacing()
                            imgui.spacing()
                            imgui.spacing()
                            imgui.spacing()
                            imgui.spacing()
        imgui.end()

    def on_gui(self) -> None:
        imgui.dock_space_over_viewport(flags=imgui.DockNodeFlags.PASSTHRU_CENTRAL_NODE)

        if self.multiviewport:
            imgui.push_style_var_im_vec2(imgui.StyleVar.WINDOW_PADDING, imgui.Vec2(0.0, 0.0))
            fb_width, fb_height = self._get_framebuffer_size()

            io = imgui.get_io()
            imgui_modifiers = io.key_mods

            mapped_mods = (
                (Modifiers.SHIFT if imgui_modifiers & imgui.Key.MOD_SHIFT else 0)
                | (Modifiers.CTRL if imgui_modifiers & imgui.Key.MOD_CTRL else 0)
                | (Modifiers.ALT if imgui_modifiers & imgui.Key.MOD_ALT else 0)
                | (Modifiers.SUPER if imgui_modifiers & imgui.Key.MOD_SUPER else 0)
            )
            mouse_button_map = {
                MouseButton.LEFT: imgui.MouseButton.LEFT,
                MouseButton.RIGHT: imgui.MouseButton.RIGHT,
                MouseButton.MIDDLE: imgui.MouseButton.MIDDLE,
            }
            rotate_button = mouse_button_map[self.key_map.camera_rotate.button]
            pan_button = mouse_button_map[self.key_map.camera_pan.button]

            prevent_left_button_window_move = (
                self.key_map.camera_rotate.button == MouseButton.LEFT
                and self.key_map.camera_rotate.mods == Modifiers.NONE
            ) or (
                self.key_map.camera_pan.button == MouseButton.LEFT and self.key_map.camera_pan.mods == Modifiers.NONE
            )

            if prevent_left_button_window_move:
                old_move_from_title_bar_only_config = io.config_windows_move_from_title_bar_only
                io.config_windows_move_from_title_bar_only = True

            output_width, output_height = self._get_framebuffer_size()

            for v in self.viewports:
                assert v.imgui_texture is not None
                assert v.image is not None

                imgui.set_next_window_size_constraints(imgui.Vec2(0, 0), imgui.Vec2(v.image.width, v.image.height))
                if imgui.begin(v.name)[0]:
                    cursor_pos = imgui.get_cursor_screen_pos()
                    pos = ivec2(cursor_pos.x, cursor_pos.y)
                    avail = imgui.get_content_region_avail()
                    size = ivec2(avail.x, avail.y)
                    v.rect.x = pos.x
                    v.rect.y = pos.y
                    v.resize(min(size.x, output_width), min(size.y, output_height))
                    imgui.image(
                        v.imgui_texture, avail, imgui.Vec2(0, 0), imgui.Vec2(size.x / fb_width, size.y / fb_height)
                    )

                    # Rotate
                    if imgui.is_item_clicked(rotate_button) and mapped_mods == self.key_map.camera_rotate.mods:
                        self.active_viewport = v
                        mouse_pos = imgui.get_mouse_pos()
                        v.on_rotate_press(ivec2(mouse_pos.x, mouse_pos.y))
                    elif not imgui.is_mouse_down(rotate_button):
                        v.on_rotate_release()

                    # Pan
                    if imgui.is_item_clicked(pan_button) and mapped_mods == self.key_map.camera_pan.mods:
                        self.active_viewport = v
                        mouse_pos = imgui.get_mouse_pos()
                        v.on_pan_press(ivec2(mouse_pos.x, mouse_pos.y))
                    elif not imgui.is_mouse_down(pan_button):
                        v.on_pan_release()

                    # Zoom
                    if imgui.is_item_hovered():
                        if mapped_mods == self.key_map.camera_zoom_modifiers:
                            v.on_zoom(dvec2(0, io.mouse_wheel))
                        elif mapped_mods == self.key_map.camera_zoom_move_modifiers:
                            v.on_zoom_with_movement(dvec2(0, io.mouse_wheel))
                else:
                    v.rect.width = 0
                    v.rect.height = 0
                imgui.end()

            if prevent_left_button_window_move:
                io.config_windows_move_from_title_bar_only = old_move_from_title_bar_only_config

            imgui.pop_style_var()

        if self.gui_show_stats:
            self.gui_stats()
        if self.gui_show_playback:
            self.gui_playback()
        if self.gui_show_inspector:
            self.gui_inspector()

    def on_gui_after_render(self) -> None:
        # Renderer GUI is drawn after because render state is updated by rendering (read-only in GUI).
        if self.gui_show_renderer:
            self.gui_renderer()

        for w in self.scene.widgets:
            w.gui()

    def run(self) -> None:
        if self.window is None:
            raise RuntimeError("Viewer.run() can only be called if a window is created (Config.window must be True)")

        self.last_frame_timestamp = perf_counter_ns()

        self.running = True
        while self.running:
            # Draw first to display window on wayland
            self.on_draw()

            # Process events
            should_wait = self.wait_events and not (self.playback.playing and self.playback.num_frames > 1)
            if self.renderer.path_tracer:
                should_wait = should_wait and not any(
                    v.path_tracer_viewport.sample_index < self.renderer.path_tracer_max_samples_per_pixel
                    for v in self.viewports
                    if v.path_tracer_viewport
                )
            process_events(should_wait)

            # Check if window was closed
            if self.window.should_close():
                break

            # Process server messages
            if self.server is not None:
                while True:
                    try:
                        client, raw_message = self.server_message_queue.get_nowait()
                    except Empty:
                        break
                    self.on_raw_message(client, raw_message)
        self.device.wait_idle()

        if self.server is not None:
            self.server.shutdown()



from __future__ import annotations

import time
from dataclasses import dataclass
from queue import Queue
from threading import Thread
from time import sleep
from typing import Callable, Optional

import numpy as np
from pyglm.glm import vec3
from pyxpg import imgui

from ambra.config import CameraConfig, Config, GuiConfig, PlaybackConfig
from ambra.primitives3d import Colormap, ColormapDistanceToPlane, ColormapKind, Points
from ambra.property import AnimationBoundary, ListBufferProperty, ListTimeSampledAnimation, UploadSettings
from ambra.viewer import Viewer


@dataclass
class CameraReading:
    camera_index: int
    data: np.ndarray


WIDTH = 320
HEIGHT = 240


class Camera:
    def __init__(
        self,
        camera_index: int,
        device_path: str,
        on_frame: Callable[[CameraReading], None],
    ):
        self.camera_index = camera_index
        self.device_path = device_path
        self.on_frame = on_frame

        self.should_stop = False
        self.thread = Thread(target=self._entry)
        self.thread.start()

    def _entry(self):
        # TODO: open camera and start streaming

        start_t = time.monotonic_ns()
        while True:
            if self.should_stop:
                break
            sleep(1.0 / 30.0)

            t = (time.monotonic_ns() - start_t) * 1e-9 * 2
            x, y = np.meshgrid(np.arange(WIDTH, dtype=np.float32), np.arange(HEIGHT, dtype=np.float32))
            x = x.flatten() / WIDTH * 2.0
            y = y.flatten() / HEIGHT * 2.0
            z = np.sin(x * 5 + t) * np.cos(y * 5 + t) * 0.5 + 0.5
            pts = np.vstack((x, y, z), dtype=np.float32).T

            self.on_frame(CameraReading(self.camera_index, pts))

    def stop(self):
        self.should_stop = True
        self.thread.join()


def main():
    queue: Queue[CameraReading] = Queue()

    def on_frame(reading: CameraReading):
        queue.put(reading)

    camera = Camera(0, "", on_frame)

    try:
        pc: Optional[Points] = None

        total_idx = 0.0
        points_animation = ListTimeSampledAnimation(AnimationBoundary.HOLD, [total_idx])

        class CustomViewer(Viewer):
            def on_draw(self):
                nonlocal pc, total_idx
                if not queue.empty():
                    reading = queue.get()
                    if pc is None:
                        if False:
                            points_property = ListBufferProperty(
                                [reading.data],
                                np.float32,
                                (-1, 3),
                                reading.data.nbytes,
                                upload=UploadSettings(batched=False),
                            )
                        else:
                            points_property = ListBufferProperty(
                                [reading.data],
                                np.float32,
                                (-1, 3),
                                reading.data.nbytes,
                                # animation=points_animation,
                                upload=UploadSettings(batched=False, preupload=False),
                            )
                        pc = Points(
                            points_property, colormap=ColormapDistanceToPlane(ColormapKind.JET, 0, 1), point_size=3
                        )
                        self.scene.objects.append(pc)
                    else:
                        total_idx += 1

                        # pc.points.append_frames([reading.data] * 10)

                        pc.points.append_frame(reading.data)
                        # points_animation.timestamps.append(float(total_idx))

                        self.playback.set_max_time(pc.points.end_animation_time(self.playback.frames_per_second))
                        if True:
                            if not self.gui_playback_slider_held:
                                self.playback.set_frame(self.playback.num_frames - 1)
                return super().on_draw()

            def on_gui(self):
                nonlocal pc

                super().on_gui()
                if imgui.begin("Editor")[0]:
                    if pc is not None:
                        u, ps = imgui.slider_float("point size", pc.point_size.get_current(), 1, 10)
                        if u:
                            pc.point_size.update_frame(0, ps)

                        if imgui.button("Remove frames"):
                            pc.points.remove_frame_range(1, pc.points.num_frames)
                            pc.points.update(self.playback.current_time, self.playback.current_frame)

                        if imgui.button("Destroy"):
                            self.scene.objects.clear()
                            pc.destroy()
                            pc = None

                imgui.end()

        viewer = CustomViewer(
            config=Config(
                camera=CameraConfig(
                    position=vec3(3),
                    target=vec3(0),
                ),
                gui=GuiConfig(stats=True, playback=True, renderer=True, inspector=True),
            ),
        )

        viewer.run()
    finally:
        camera.stop()


main()



import numpy as np
from pyxpg import imgui
from pyxpg.imgui import implot

from ambra.config import Config, GuiConfig, RendererConfig
from ambra.utils.hook import hook
from ambra.viewer import Viewer

data = np.array(
    [
        [1, 5, 3, 2],
        [10, 50, 30, 20],
        [100, 500, 300, 200],
    ],
    np.int64,
)


class CustomViewer(Viewer):
    @hook
    def on_gui(self):
        if imgui.begin("YO")[0]:
            if implot.begin_plot("Plot", flags=implot.PlotFlags.NO_FRAME):
                implot.plot_line("Line", data[:, 0], data[:, 1][::-1])
                implot.end_plot()
        imgui.end()


viewer = CustomViewer(
    config=Config(
        preferred_frames_in_flight=3,
        gui=GuiConfig(
            stats=True,
        ),
    ),
)

viewer.run()



import numpy as np

from ambra.config import (
    CameraConfig,
    CameraProjection,
    Config,
    GuiConfig,
    PlaybackConfig,
    RendererConfig,
    UploadMethod,
)
from ambra.primitives2d import Lines
from ambra.property import ArrayBufferProperty, ListBufferProperty, UploadSettings
from ambra.viewer import Viewer

viewer = Viewer(
    config=Config(
        window_width=1920,
        window_height=1080,
        playback=PlaybackConfig(
            playing=True,
        ),
        camera=CameraConfig(
            projection=CameraProjection.ORTHOGRAPHIC,
            ortho_half_extents=(10, 10),
            position=(0, 0, 0),
            target=(0, 0, 1),
        ),
        world_up=(0, 1, 0),
        renderer=RendererConfig(
            force_buffer_upload_method=UploadMethod.GRAPHICS_QUEUE,
        ),
        gui=GuiConfig(
            stats=True,
            renderer=True,
        ),
    ),
)

positions = np.linspace(
    np.array(
        [
            [0.0, 0.0],
            [1.0, 0.0],
            [0.0, 0.0],
            [0.0, 1.0],
            [0.0, 0.0],
            [1.0, 1.0],
        ],
        np.float32,
    ),
    np.array(
        [
            [0.0, 0.0],
            [1.0, 0.0],
            [0.0, 0.0],
            [0.0, 1.0],
            [0.0, 0.0],
            [1.0, 1.0],
        ],
        np.float32,
    )
    * 10,
    100,
)

colors = [
    np.array(
        [
            0xFF0000FF,
            0xFF0000FF,
            0xFF00FF00,
            0xFF00FF00,
            0xFFFF0000,
            0xFFFF0000,
        ],
        np.uint32,
    ),
    np.array(
        [
            0xFF0000FF,
            0xFF0000FF,
            0xFF00FF00,
            0xFF00FF00,
            0xFFFF0000,
            0xFFFF0000,
        ],
        np.uint32,
    ),
]

positions = ArrayBufferProperty(positions, np.float32, upload=UploadSettings(preupload=False, batched=True))
colors = ListBufferProperty(colors, shape=(-1,), upload=UploadSettings(preupload=True, batched=False))

line = Lines(positions, colors, 4)
viewer.scene.objects.extend([line])

viewer.run()



# Copyright Dario Mylonopoulos
# SPDX-License-Identifier: MIT

from dataclasses import dataclass
from typing import Any, Optional

import numpy as np
from numpy.typing import NDArray
from pyxpg import (
    AllocType,
    Buffer,
    BufferUsageFlags,
    CommandBuffer,
    Device,
    DeviceFeatures,
    Fence,
    Format,
    Image,
    ImageLayout,
    ImageUsageFlags,
    MemoryUsage,
)

from .renderer import FrameInputs
from .utils.gpu import _format_to_channels_dtype_int_bgra_table, get_image_pitch_rows_and_texel_size
from .utils.ring_buffer import RingBuffer


@dataclass
class HeadlessSwapchainFrame:
    image: Optional[Image]
    readback_buffer: Optional[Buffer]
    command_buffer: CommandBuffer
    transfer_command_buffer: Optional[CommandBuffer]
    fence: Fence

    def issue_readback(self) -> None:
        assert self.readback_buffer is not None
        assert self.image is not None

        self.command_buffer.image_barrier(
            self.image, ImageLayout.TRANSFER_SRC_OPTIMAL, MemoryUsage.NONE, MemoryUsage.TRANSFER_SRC
        )
        self.command_buffer.copy_image_to_buffer(self.image, self.readback_buffer)
        self.command_buffer.memory_barrier(MemoryUsage.TRANSFER_SRC, MemoryUsage.HOST_READ)

    def realize_readback(self) -> NDArray[Any]:
        assert self.readback_buffer is not None
        assert self.image is not None

        self.fence.wait()

        channels, dtype, _, bgra = _format_to_channels_dtype_int_bgra_table[self.image.format]
        shape = (self.image.height, self.image.width, channels)
        img = np.frombuffer(self.readback_buffer.data, dtype).copy().reshape(shape)
        if bgra:
            img[:, :, (0, 2)] = img[:, :, (2, 0)]
        return img


class HeadlessSwapchain:
    def __init__(self, device: Device, num_frames_in_flight: int, format: Format):
        self.device = device
        self.num_frames_in_flight = num_frames_in_flight
        self.format = format
        self.width = 0
        self.height = 0

        self.frames = RingBuffer(
            [
                HeadlessSwapchainFrame(
                    image=None,
                    readback_buffer=None,
                    command_buffer=CommandBuffer(device, name=f"windowless-swapchain-commands-{i}"),
                    transfer_command_buffer=CommandBuffer(
                        device, device.transfer_queue_family_index, name=f"windowless-swapchain-transfer-commands-{i}"
                    )
                    if self.device.has_transfer_queue
                    else None,
                    fence=Fence(device, signaled=True, name=f"windowless-swapchain-fence-{i}"),
                )
                for i in range(num_frames_in_flight)
            ]
        )

    def ensure_size(self, width: int, height: int) -> None:
        if self.width == width and self.height == height:
            return

        pitch, rows, _ = get_image_pitch_rows_and_texel_size(width, height, self.format)
        size = pitch * rows

        usage_flags = ImageUsageFlags.TRANSFER_SRC | ImageUsageFlags.TRANSFER_DST | ImageUsageFlags.COLOR_ATTACHMENT
        if self.device.features & DeviceFeatures.RAY_QUERY:
            usage_flags |= ImageUsageFlags.STORAGE

        for i, f in enumerate(self.frames):
            f.image = Image(
                self.device,
                width,
                height,
                self.format,
                usage_flags,
                AllocType.DEVICE,
                name=f"windowless-swapchain-image-{i}",
            )
            f.readback_buffer = Buffer(
                self.device,
                size,
                BufferUsageFlags.TRANSFER_DST | BufferUsageFlags.TRANSFER_SRC,
                AllocType.HOST,
                name=f"windowless-swapchain-buffer-{i}",
            )

    def begin_frame(self) -> FrameInputs:
        frame = self.frames.get_current()
        assert frame.readback_buffer is not None
        assert frame.image is not None

        frame.fence.wait_and_reset()
        return FrameInputs(
            frame.image,
            frame.command_buffer,
            frame.transfer_command_buffer,
            [],
            [],
        )

    def end_frame(self, frame_inputs: FrameInputs) -> None:
        frame = self.frames.get_current()
        self.device.graphics_queue.submit(
            frame.command_buffer,
            wait_semaphores=[(s.sem, s.wait_value, s.wait_stage) for s in frame_inputs.additional_semaphores],
            signal_semaphores=[(s.sem, s.signal_value, s.signal_stage) for s in frame_inputs.additional_semaphores],
            fence=frame.fence,
        )

    def advance(self) -> None:
        self.frames.advance()

    def get_current(self) -> HeadlessSwapchainFrame:
        return self.frames.get_current()

    def get_current_and_advance(self) -> HeadlessSwapchainFrame:
        return self.frames.get_current_and_advance()



import numpy as np
from pyglm.glm import inverse, ivec2, normalize, quatLookAtRH, vec3, vec4
from pyxpg import *

from ambra.config import CameraConfig, Config, GuiConfig, PlaybackConfig, RendererConfig
from ambra.geometry import create_axis3d_lines_and_colors, create_cube, create_plane
from ambra.lights import DirectionalLight, DirectionalShadowSettings
from ambra.materials import DiffuseMaterial, PBRMaterial
from ambra.primitives3d import Lines, Mesh
from ambra.utils.descriptors import create_descriptor_layout_pool_and_set
from ambra.utils.hook import hook
from ambra.viewer import Viewer


class CustomViewer(Viewer):
    def __init__(self, title="ambra", config=None, key_map=None):
        super().__init__(title, config, key_map)

    @hook
    def on_gui(self):
        for light in [light1, light2]:
            if imgui.begin(light.name)[0]:
                if hasattr(light, "shadow_map"):
                    u, t = imgui.drag_float3("Translation", light.translation.get_current(), 0.01)
                    if u:
                        light.translation.update_frame(0, np.array(t))

                    u, s = imgui.drag_float3("Scale", light.scale.get_current(), 0.01)
                    if u:
                        light.scale.update_frame(0, np.array(s))

                    u, r = imgui.drag_float4("Rotation", light.rotation.get_current(), 0.1)

                    _, light.shadow_settings.min_bias = imgui.drag_float(
                        "Min Bias", light.shadow_settings.min_bias, 0.0001, 0, 1.0, "%.4f"
                    )
                    _, light.shadow_settings.max_bias = imgui.drag_float(
                        "Max Bias", light.shadow_settings.max_bias, 0.0001, 0, 1.0, "%.4f"
                    )

                    if u:
                        light.rotation.update_frame(0, np.array(normalize(vec4(r))))
                    if light._texture is None and light.shadow_map is not None:
                        layout, pool, set = create_descriptor_layout_pool_and_set(
                            viewer.device,
                            [
                                DescriptorSetBinding(
                                    1, DescriptorType.COMBINED_IMAGE_SAMPLER, stage_flags=Stage.FRAGMENT
                                ),
                            ],
                        )
                        set.write_combined_image_sampler(
                            light.shadow_map, ImageLayout.SHADER_READ_ONLY_OPTIMAL, sampler, 0
                        )
                        light._texture = imgui.Texture(set)

                    if light._texture is not None:
                        avail = imgui.get_content_region_avail()
                        available = ivec2(avail.x, avail.y)

                        ar = 1.0

                        # height = available.x / ar
                        height = available.y
                        view_size = ivec2(ar * height, height)
                        imgui.image(
                            light._texture,
                            imgui.Vec2(*view_size),
                            uv0=(0, 0),
                            uv1=(1, -1),
                        )

            imgui.end()


viewer = CustomViewer(
    config=Config(
        renderer=RendererConfig(
            msaa_samples=4,
            path_tracer_max_bounces=0,
        ),
        playback=PlaybackConfig(
            playing=False,
        ),
        camera=CameraConfig(
            position=vec3(10, 4, 10),
            target=vec3(0),
        ),
        world_up=(0, 0, 1),
        gui=GuiConfig(
            stats=True,
            # multiviewport=True,
            # initial_number_of_viewports=2,
            inspector=True,
        ),
    ),
)

translation = np.linspace((0, 0, 0.0), (5, 5, 0), num=50)
instance_positions = np.array([(-2, 0, 0), (2, 0, 0)])

# material = PBRMaterial((1, 1, 1), 1.0, 0.0)
material = DiffuseMaterial((1, 1, 1))

cube_positions, cube_normals, cube_indices = create_cube()
m = Mesh(
    cube_positions,
    cube_indices,
    translation=translation,
    cull_mode=CullMode.BACK,
    front_face=FrontFace.COUNTER_CLOCKWISE,
    instance_positions=instance_positions,
    material=material,
)

plane_positions, plane_normals, plane_indices = create_plane((0, 0, -0.51), (10, 10))
p = Mesh(plane_positions, plane_indices, plane_normals, material=material)
o = Lines(*create_axis3d_lines_and_colors(), 4.0)

light1 = DirectionalLight.look_at(
    # vec3(-2, 3, 4),
    vec3(2, 3, 4),
    vec3(0, 0, 0),
    vec3(0, 0, 1),
    np.array([1.0, 1.0, 1.0]),
    shadow_settings=DirectionalShadowSettings(half_extent=5.0, z_near=1.0, z_far=10),
)
light1._texture = None

light2 = DirectionalLight.look_at(
    vec3(-2, 3, 4),
    vec3(0, 0, 0),
    vec3(0, 0, 1),
    np.array([1.0, 1.0, 1.0]),
    shadow_settings=DirectionalShadowSettings(half_extent=5.0, z_near=1.0, z_far=10),
)
light2._texture = None

sampler = Sampler(
    viewer.device,
    u=SamplerAddressMode.REPEAT,
    v=SamplerAddressMode.REPEAT,
)

viewer.scene.objects.extend(
    [
        m,
        p,
        o,
        light1,
        light2,
    ]
)

viewer.run()



# Copyright Dario Mylonopoulos
# SPDX-License-Identifier: MIT

from dataclasses import dataclass, field
from enum import Enum
from typing import Optional, Tuple

from pyxpg import DefaultFontPreference, LogLevel


@dataclass
class ServerConfig:
    enabled: bool = False
    address: str = "localhost"
    port: int = 9168


class UploadMethod(Enum):
    GRAPHICS_QUEUE = 0
    TRANSFER_QUEUE = 1
    MAPPED_PREFER_HOST = 2
    MAPPED_PREFER_DEVICE = 3


class RenderMode(Enum):
    RASTER = 0
    PATH_TRACER = 1


@dataclass
class RendererConfig:
    background_color: Tuple[float, float, float, float] = (1, 1, 1, 1)
    render_mode: RenderMode = RenderMode.RASTER
    path_tracer_max_bounces: int = 4
    path_tracer_samples_per_frame: int = 4
    path_tracer_max_samples_per_pixel: int = 1024
    path_tracer_max_textures: int = 32 * 1024
    path_tracer_clip_value: float = 10.0
    path_tracer_use_background_color: bool = True
    msaa_samples: int = 1
    uniform_pool_block_size: int = 4 * 1024 * 1024
    upload_buffer_size: int = 32 * 1024 * 1024
    upload_buffer_count: int = 2
    thread_pool_workers: Optional[int] = None
    use_transfer_queue_if_available: bool = True
    force_buffer_upload_method: Optional[UploadMethod] = None
    force_image_upload_method: Optional[UploadMethod] = None
    max_lights_per_type: int = 32
    max_shadow_maps: int = 8
    mip_generation_batch_size: int = 32


@dataclass
class PlaybackConfig:
    playing: bool = False
    frames_per_second: float = 30.0
    playback_speed_multiplier: float = 1.0
    initial_time: float = 0.0
    initial_frame: Optional[int] = None
    max_time: Optional[float] = None


class CameraProjection(Enum):
    PERSPECTIVE = 0
    ORTHOGRAPHIC = 1


class CameraControlMode(Enum):
    NONE = 0
    ORBIT = 1
    TRACKBALL = 2
    FIRST_PERSON = 3
    # PAN_AND_ZOOM_ORTHO = 4


class Handedness(Enum):
    RIGHT_HANDED = 0
    LEFT_HANDED = 1


@dataclass
class GuiConfig:
    multiviewport: bool = False
    initial_number_of_viewports: int = 1
    max_viewport_count: int = 8

    ini_filename: Optional[str] = "imgui.ini"  # if set to None disables.

    default_font_size: Optional[float] = None  # If None use default imgui font size (13.0)
    default_font_preference: DefaultFontPreference = (
        DefaultFontPreference.AUTO
    )  # If set to AUTO use BITMAP for font size < 15.0 and VECTOR otherwise.
    font_scale: float = (
        1.0  # Scale all fonts globally by this factor (default font selection does not take this factor into account).
    )

    stats: bool = False
    playback: bool = False
    inspector: bool = False
    renderer: bool = False


@dataclass
class CameraConfig:
    # Inital state
    projection: CameraProjection = CameraProjection.PERSPECTIVE
    position: Tuple[float, float, float] = (1.0, 1.0, 1.0)
    target: Tuple[float, float, float] = (0.0, 0.0, 0.0)
    z_near: float = 0.01
    z_far: float = 1000.0

    # If type is CameraProjection.PERSPECTIVE
    perspective_vertical_fov: float = 45.0

    # If type is CameraProjection.ORTHOGRAPHIC
    ortho_center: Tuple[float, float] = (0.0, 0.0)
    ortho_half_extents: Tuple[float, float] = (1.0, 1.0)

    # Viewport camera controls
    control_mode: CameraControlMode = CameraControlMode.ORBIT
    rotation_speed: Tuple[float, float] = (0.005, 0.005)
    pan_speed: Tuple[float, float] = (0.01, 0.01)
    pan_distance_speed_scale: float = 0.1
    pan_min_speed_scale: float = 0.1
    zoom_speed: float = 0.1
    zoom_distance_speed_scale: float = 1.0
    zoom_min_speed_scale: float = 2.0
    zoom_min_target_distance: float = 0.01
    zoom_ortho_speed: float = 0.04


@dataclass
class Config:
    # Logging.
    #
    # If not None, the Viewer constructor sets the pyxpg log level to this value.
    # If None preserve the existing pyxpg log level.
    # The pyxpg log level is global for the module and defaults to DISABLED.
    #
    # Log messages from pyxpg are always printed to stdout.
    log_level: Optional[LogLevel] = None

    # Window
    window: bool = True
    window_x: Optional[int] = None
    window_y: Optional[int] = None
    window_width: int = 1280
    window_height: int = 720
    wait_events: bool = False
    vsync: bool = True
    preferred_frames_in_flight: int = 2

    # Vulkan
    force_physical_device_index: Optional[int] = None
    prefer_discrete_gpu: bool = True
    enable_validation_layer: bool = False
    enable_synchronization_validation: bool = False
    enable_gpu_based_validation: bool = False
    enable_shader_debug_printf: bool = False
    shader_debug_printf_buffer_size: int = 4096

    # Scene
    world_up: Tuple[float, float, float] = (0, 0, 1)
    handedness: Handedness = Handedness.RIGHT_HANDED

    # Stats
    stats_frame_time_count: int = 32

    # Sub-configs
    camera: CameraConfig = field(default_factory=CameraConfig)
    renderer: RendererConfig = field(default_factory=RendererConfig)
    playback: PlaybackConfig = field(default_factory=PlaybackConfig)
    server: ServerConfig = field(default_factory=ServerConfig)
    gui: GuiConfig = field(default_factory=GuiConfig)



from typing import Optional, Tuple

import numpy as np
from pyglm.glm import vec3

from .lights import DirectionalLight, DirectionalShadowSettings, UniformEnvironmentLight


def two_directional_lights_and_uniform_environment_light(
    front_position: Tuple[float, float, float] = (2, 3, 4),
    front_target: Tuple[float, float, float] = (0, 0, 0),
    front_world_up: Tuple[float, float, float] = (0, 0, 1),
    front_radiance: Tuple[float, float, float] = (1, 1, 1),
    front_shadow_settings: Optional[DirectionalShadowSettings] = None,
    back_position: Tuple[float, float, float] = (-2, -3, 4),
    back_target: Tuple[float, float, float] = (0, 0, 0),
    back_world_up: Tuple[float, float, float] = (0, 0, 1),
    back_radiance: Tuple[float, float, float] = (0.75, 0.75, 0.75),
    back_shadow_settings: Optional[DirectionalShadowSettings] = None,
    uniform_radiance: Tuple[float, float, float] = (0.2, 0.2, 0.2),
) -> Tuple[DirectionalLight, DirectionalLight, UniformEnvironmentLight]:
    front = DirectionalLight.look_at(
        vec3(front_position),
        vec3(front_target),
        vec3(front_world_up),
        np.array(front_radiance),  # type: ignore
        shadow_settings=front_shadow_settings or DirectionalShadowSettings(half_extent=5.0, z_near=1.0, z_far=10),
        name="Front light",
    )

    back = DirectionalLight.look_at(
        vec3(back_position),
        vec3(back_target),
        vec3(back_world_up),
        np.array(back_radiance),  # type: ignore
        shadow_settings=back_shadow_settings or DirectionalShadowSettings(half_extent=5.0, z_near=1.0, z_far=10),
        name="Back light",
    )

    uniform = UniformEnvironmentLight(uniform_radiance, name="Uniform light")  # type: ignore

    return front, back, uniform



# Copyright Dario Mylonopoulos
# SPDX-License-Identifier: MIT

from dataclasses import dataclass
from enum import Enum, auto
from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np
from numpy.typing import NDArray
from pyxpg import (
    AllocType,
    Buffer,
    BufferUsageFlags,
    CommandBuffer,
    CullMode,
    Device,
    DeviceFeatures,
    Fence,
    Format,
    Image,
    ImageLayout,
    ImageUsageFlags,
    ImageView,
    MemoryUsage,
    Stage,
    get_format_info,
)

from .ring_buffer import RingBuffer


def is_pow_2(n: int) -> int:
    return n != 0 and (n & (n - 1) == 0)


def align_up(v: int, a: int) -> int:
    return (v + a - 1) & ~(a - 1)


def div_round_up(a: int, b: int) -> int:
    return (a + b - 1) // b


def view_bytes(a: NDArray[Any]) -> memoryview:
    return a.reshape((-1,)).view(np.uint8).data


@dataclass
class BufferUploadInfo:
    data: memoryview
    buffer: Buffer
    offset: int


class MipGenerationFilter(Enum):
    AVERAGE = auto()
    AVERAGE_SRGB = auto()
    MAX = auto()
    MIN = auto()


@dataclass
class ImageUploadInfo:
    data: memoryview
    image: Image
    layout: ImageLayout

    level_0_view: Optional[ImageView]  # If None, no mip generation will happen
    mip_views: List[ImageView]
    mip_generation_filter: MipGenerationFilter


@dataclass
class AccelerationStructureInstanceInfo:
    transform: NDArray[np.float32]
    normal_matrix: NDArray[np.float32]

    positions_count: int
    positions_address: int
    normals_address: int
    tangents_address: int
    uvs_address: int

    primitive_count: int
    indices_address: int

    material_index: int
    viewport_mask: int


@dataclass
class BulkUploadState:
    buffer: Buffer
    cmd: CommandBuffer
    fence: Fence
    submitted: bool


class BulkUploader:
    def __init__(self, device: Device, size: int, count: int):
        self.device = device

        self.upload_states: List[BulkUploadState] = []
        upload_alignment = max(
            device.device_properties.limits.optimal_buffer_copy_offset_alignment,
            device.device_properties.limits.optimal_buffer_copy_row_pitch_alignment,
        )

        if device.has_transfer_queue:
            self.queue = device.transfer_queue
            self.queue_family_index = device.transfer_queue_family_index
        else:
            self.queue = device.graphics_queue
            self.queue_family_index = device.graphics_queue_family_index

        for i in range(count):
            self.upload_states.append(
                BulkUploadState(
                    buffer=Buffer(
                        device,
                        align_up(size, upload_alignment),
                        BufferUsageFlags.TRANSFER_SRC,
                        AllocType.HOST,
                        name=f"bulk-upload-buffer-{i}",
                    ),
                    cmd=CommandBuffer(device, self.queue_family_index, name=f"bulk-upload-commands-{i}"),
                    fence=Fence(device, name=f"bulk-upload-fence-{i}"),
                    submitted=False,
                )
            )

    def bulk_upload(
        self, uploads: List[Union[BufferUploadInfo, ImageUploadInfo]], mip_generation_requests: List[ImageUploadInfo]
    ) -> None:
        # Compute alignment requirements
        offset_alignment = max(self.device.device_properties.limits.optimal_buffer_copy_offset_alignment, 16)
        pitch_alignment = max(self.device.device_properties.limits.optimal_buffer_copy_row_pitch_alignment, 16)

        # State of current upload
        start_image_plane = 0
        start_image_row = 0
        start_buffer_offset = 0

        # Stats
        total_bytes_uploaded = 0
        # print(f"Bulk uploads: {len(uploads)}")
        # begin_timestamp = perf_counter_ns()

        i = 0
        upload_buffer_index = 0
        while i < len(uploads):
            # print(f"Upload index {upload_buffer_index:3}")
            state = self.upload_states[upload_buffer_index]

            if state.submitted:
                state.fence.wait_and_reset()
                state.submitted = False

            with state.cmd as cmd:
                offset = 0

                while i < len(uploads) and offset < state.buffer.size:
                    state = self.upload_states[upload_buffer_index]

                    # Space left in upload buffer
                    remaining_size = state.buffer.size - offset

                    info = uploads[i]

                    if isinstance(info, BufferUploadInfo):
                        total_size = len(info.data)
                        size = total_size - start_buffer_offset
                        fitting_size = min(remaining_size, size)

                        # print(f"    Buffer {i:3} - offset {offset:12} - size {size:12} - remaining_size {remaining_size:12} - fitting size {fitting_size:12} - bufoffset {start_buffer_offset:12}")
                        state.buffer.data[offset : offset + fitting_size] = info.data[
                            start_buffer_offset : start_buffer_offset + fitting_size
                        ]

                        cmd.copy_buffer_range(
                            state.buffer, info.buffer, fitting_size, offset, info.offset + start_buffer_offset
                        )

                        if start_buffer_offset + fitting_size == total_size:
                            start_buffer_offset = 0
                            i += 1
                            total_bytes_uploaded += total_size
                        else:
                            start_buffer_offset += fitting_size
                    elif isinstance(info, ImageUploadInfo):
                        input_pitch, total_rows, texel_size = get_image_pitch_rows_and_texel_size(
                            info.image.width, info.image.height, info.image.format
                        )
                        total_planes = info.image.depth
                        if input_pitch * total_rows * total_planes != len(info.data):
                            # print(type(info.data), info.data.shape)
                            raise ValueError(
                                f"ImageUploadInfo data size ({len(info.data)}) does not match pitch and rows ({input_pitch} x {total_rows} x {total_planes} = {input_pitch * total_rows * total_planes})"
                            )

                        # Adjust rows by number of rows already uploaded and align pitch
                        rows = total_rows - start_image_row
                        pitch = align_up(input_pitch, pitch_alignment)
                        plane_size = pitch * rows
                        planes = total_planes - start_image_plane

                        if pitch > state.buffer.size:
                            raise RuntimeError(
                                f"Image pitch ({pitch}) is larger than upload buffer size ({state.buffer.size})"
                            )

                        # Stop if we can't fit a single row
                        if pitch > remaining_size:
                            break

                        # Compute how many rows we are actually uploading in this command

                        # Do 2D copy if we are in middle of a plane or less than one plane fits in the buffer
                        if start_image_row != 0 or remaining_size <= plane_size:
                            fitting_rows = min(rows, remaining_size // pitch)
                            fitting_planes = 1
                        else:
                            fitting_rows = total_rows
                            fitting_planes = min(planes, remaining_size // plane_size)

                        fitting_size = fitting_planes * fitting_rows * pitch
                        # print(f"    Image  {i:3} - offset {offset:12} - pitch {pitch:12} - rows {rows:12} - planes {planes:12} - remaining_size {remaining_size:12} - fitting_rows {fitting_rows:12} - fitting planes {fitting_planes:12} - fitting_size {fitting_size:12} - start_image_row {start_image_row:12} - start_image_plane {start_image_plane} - pitch alignment {pitch_alignment}")

                        # Copy image data to staging buffer expanding to correct pitch
                        upload_buffer_range = state.buffer.data[offset : offset + fitting_size]
                        # print(upload_buffer_range.shape, info.data.shape)
                        upload_buffer_3d_view = np.frombuffer(upload_buffer_range, np.uint8).reshape(
                            (fitting_planes, fitting_rows, pitch)
                        )
                        data_buffer_3d_view = np.frombuffer(info.data, np.uint8).reshape(
                            (total_planes, total_rows, input_pitch)
                        )

                        # print(upload_buffer_3d_view.shape, data_buffer_3d_view.shape)

                        # TODO: correctly handle block formats here, need to think about data shape looks for those
                        upload_buffer_3d_view[:, :input_pitch] = data_buffer_3d_view[
                            start_image_plane : start_image_plane + fitting_planes,
                            start_image_row : start_image_row + fitting_rows,
                            :input_pitch,
                        ]

                        # Tranisition to TRANSFER_DST_OPTIMAL if first upload
                        if start_image_plane == 0 and start_image_row == 0:
                            cmd.image_barrier(
                                info.image,
                                ImageLayout.TRANSFER_DST_OPTIMAL,
                                MemoryUsage.NONE,
                                MemoryUsage.TRANSFER_DST,
                                undefined=True,
                            )

                        # Upload to image range
                        cmd.copy_buffer_to_image_range(
                            state.buffer,
                            info.image,
                            info.image.width,
                            fitting_rows,
                            0,
                            start_image_row,
                            offset,
                            pitch // texel_size,
                            0,
                            fitting_planes,
                            start_image_plane,
                        )

                        if (
                            start_image_plane + fitting_planes == total_planes
                            and start_image_row + fitting_rows == total_rows
                        ):
                            # Transition to final layout if last upload
                            final_layout = info.layout

                            # Register for mip generation if requested and transition to GENERAL layout instead
                            if info.level_0_view is not None:
                                mip_generation_requests.append(info)
                                final_layout = ImageLayout.GENERAL

                            cmd.image_barrier(info.image, final_layout, MemoryUsage.TRANSFER_DST, MemoryUsage.NONE)
                            i += 1
                            start_image_plane = 0
                            start_image_row = 0
                            total_bytes_uploaded += total_rows * pitch
                        elif start_image_row + fitting_rows == total_rows:
                            start_image_row = 0
                            start_image_plane += fitting_planes
                            # If we are uploading full planes and the image is not done break to not upload a partial plane
                            if start_image_plane < total_planes:
                                break
                        else:
                            start_image_row += fitting_rows
                    else:
                        raise TypeError(f"Expected BufferUploadInfo or ImageUploadInfo. Got {type(info)}")

                    # Advance image and buffer offset
                    offset += align_up(fitting_size, offset_alignment)

            # Submit batch signaling fence
            self.queue.submit(state.cmd, fence=state.fence)
            state.submitted = True

            # Advance buffer
            upload_buffer_index = (upload_buffer_index + 1) % len(self.upload_states)

        for s in self.upload_states:
            if s.submitted:
                s.fence.wait_and_reset()
                s.submitted = False

        # elapsed = (perf_counter_ns() - begin_timestamp) * 1e-9
        # print(f"Bulk upload of {total_bytes_uploaded/(1024 * 1024):.3f}MB in {elapsed * 1e3:6.3f}ms ({total_bytes_uploaded / (1024 * 1024 * 1024) / elapsed:.3f}GB/s)")


class UploadableBuffer(Buffer):
    def __init__(
        self,
        device: Device,
        size: int,
        usage_flags: BufferUsageFlags,
        name: Optional[str] = None,
    ):
        # TODO: we likely want an option to ensure we do not use BAR even if available for cases
        # where we want the GPU to do the upload.
        #
        # Maybe we should also handle integrated and CPU differently here
        super().__init__(
            device,
            size,
            usage_flags | BufferUsageFlags.TRANSFER_DST,
            AllocType.DEVICE_MAPPED_WITH_FALLBACK,
            name,
        )
        if not self.is_mapped:
            if name is not None:
                name = f"{name} - staging"
            self._staging = Buffer(
                device,
                size,
                BufferUsageFlags.TRANSFER_SRC,
                AllocType.HOST_WRITE_COMBINING,
                name,
            )

    @classmethod
    def from_data(  # type: ignore
        cls,
        device: Device,
        data: memoryview,
        usage_flags: BufferUsageFlags,
        name: Optional[str] = None,
    ) -> "UploadableBuffer":
        buf = cls(device, len(data), usage_flags, name=name)
        buf.upload_sync(data)
        return buf

    def upload(
        self,
        cmd: CommandBuffer,
        usage: MemoryUsage,
        data: Union[memoryview, NDArray[np.uint8]],
        offset: int = 0,
    ) -> None:
        if self.is_mapped:
            # print(self.data.c_contiguous, data.c_contiguous )
            # print(self.data.shape       , data.shape        )
            # print(self.data.readonly    , data.readonly     )
            # print(self.data.format    , data.format     )
            self.data[offset : offset + len(data)] = data
        else:
            self._staging.data[offset : offset + len(data)] = data
            cmd.copy_buffer_range(
                self._staging,
                self,
                len(data),
                src_offset=offset,
                dst_offset=offset,
            )
            if usage != MemoryUsage.NONE:
                cmd.memory_barrier(MemoryUsage.TRANSFER_DST, usage)

    def upload_sync(self, data: Union[memoryview, NDArray[np.uint8]], offset: int = 0) -> None:
        if self.is_mapped:
            # print(self.data.c_contiguous, data.c_contiguous )
            # print(self.data.shape       , data.shape        )
            # print(self.data.readonly    , data.readonly     )
            # print(self.data.format    , data.format     )
            self.data[offset : offset + len(data)] = data
        else:
            with self.device.sync_commands() as cmd:
                self.upload(cmd, MemoryUsage.NONE, data, offset)


def div_ceil(n: int, d: int) -> int:
    return (n + d - 1) // d


def get_image_pitch_rows_and_texel_size(width: int, height: int, format: Format) -> Tuple[int, int, int]:
    info = get_format_info(format)
    if info.size_of_block_in_bytes > 0:
        size = info.size_of_block_in_bytes
        pitch = div_ceil(width, info.block_side_in_pixels) * size
        rows = div_ceil(height, info.block_side_in_pixels)
    else:
        size = info.size
        pitch = width * size
        rows = height
    return pitch, rows, size


# NOTE: in the future this could use VK_EXT_host_image_copy to do uploads / transitions if available
class UploadableImage(Image):
    def __init__(
        self,
        device: Device,
        width: int,
        height: int,
        format: Format,
        usage_flags: ImageUsageFlags,
        dedicated_alloc: bool = False,
        name: Optional[str] = None,
    ):
        pitch, rows, _ = get_image_pitch_rows_and_texel_size(width, height, format)
        self._staging = Buffer(
            device,
            pitch * rows,
            BufferUsageFlags.TRANSFER_SRC,
            AllocType.HOST_WRITE_COMBINING,
            f"{name} - staging",
        )
        super().__init__(
            device,
            width,
            height,
            format,
            usage_flags | ImageUsageFlags.TRANSFER_DST,
            AllocType.DEVICE_DEDICATED if dedicated_alloc else AllocType.DEVICE,
        )

    @classmethod
    def from_data(  # type: ignore
        cls,
        device: Device,
        data: memoryview,
        layout: ImageLayout,
        width: int,
        height: int,
        format: Format,
        usage_flags: ImageUsageFlags,
        dedicated_alloc: bool = False,
        name: Optional[str] = None,
    ) -> "UploadableImage":
        buf = cls(device, width, height, format, usage_flags, dedicated_alloc, name)
        buf.upload_sync(data, layout)
        return buf

    def upload(
        self,
        cmd: CommandBuffer,
        layout: ImageLayout,
        src_usage: MemoryUsage,
        dst_usage: MemoryUsage,
        data: Union[memoryview, NDArray[np.uint8]],
    ) -> None:
        # Upload to staging buffer
        if data.shape is None or len(data.shape) != 1:
            raise ValueError(f"data must be flat array. Got shape: {data.shape}")
        if len(data) != self._staging.size:
            raise ValueError(f"data must be of size {self._staging.size}. Got size: {len(data)}")
        self._staging.data[:] = data
        cmd.image_barrier(
            self,
            ImageLayout.TRANSFER_DST_OPTIMAL,
            src_usage,
            MemoryUsage.TRANSFER_DST,
            undefined=True,
        )
        cmd.copy_buffer_to_image(self._staging, self)
        if layout == ImageLayout.UNDEFINED or layout == ImageLayout.TRANSFER_DST_OPTIMAL:
            cmd.memory_barrier(MemoryUsage.TRANSFER_DST, dst_usage)
        else:
            cmd.image_barrier(self, layout, MemoryUsage.TRANSFER_DST, dst_usage)

    def upload_sync(self, data: Union[memoryview, NDArray[np.uint8]], layout: ImageLayout) -> None:
        with self.device.sync_commands() as cmd:
            self.upload(cmd, layout, MemoryUsage.NONE, MemoryUsage.NONE, data)


@dataclass
class UniformBlockAllocation:
    buffer: UploadableBuffer
    offset: int
    size: int

    def upload(self, cmd: CommandBuffer, data: Union[memoryview, NDArray[np.uint8]]) -> None:
        if len(data) > self.size:
            raise IndexError("data is larger than buffer allocation")
        self.buffer.upload(cmd, MemoryUsage.SHADER_UNIFORM, data, self.offset)


@dataclass
class UniformBlock:
    buffers: RingBuffer[UploadableBuffer]
    size: int
    used: int = 0

    def alloc(self, size: int, alignment: int) -> UniformBlockAllocation:
        assert self.used + size < self.size
        alloc = UniformBlockAllocation(
            self.buffers.get_current(),
            self.used,
            size,
        )
        self.used += (size + alignment - 1) & ~(alignment - 1)
        return alloc

    def advance(self) -> None:
        self.buffers.advance()
        self.used = 0


class UniformPool:
    def __init__(self, device: Device, num_frames: int, block_size: int):
        self.device = device
        self.num_frames = num_frames
        self.block_size = block_size
        self.blocks: List[UniformBlock] = []
        self.alignment = max(
            16,
            self.device.device_properties.limits.min_uniform_buffer_offset_alignment,
        )
        self.max_uniform_buffer_range = self.device.device_properties.limits.max_uniform_buffer_range

        # Warmup first block
        self._alloc_block(block_size)

    def _alloc_block(self, min_size: int) -> UniformBlock:
        size = max(min_size, self.block_size)
        block_idx = len(self.blocks)
        block = UniformBlock(
            buffers=RingBuffer(
                [
                    UploadableBuffer(
                        self.device,
                        size,
                        BufferUsageFlags.UNIFORM,
                        name=f"uniform-pool-block-buf{block_idx}",
                    )
                    for _ in range(self.num_frames)
                ]
            ),
            size=size,
        )

        # Sync ringbuffer index, not necessary but makes for easier debugging
        if self.blocks:
            index = self.blocks[0].buffers.index
            block.buffers.set(index)

        self.blocks.append(block)
        return block

    def alloc(self, size: int) -> UniformBlockAllocation:
        if size > self.max_uniform_buffer_range:
            raise RuntimeError(
                f"allocation size ({size}) is larger than device limit ({self.max_uniform_buffer_range})"
            )

        for b in self.blocks:
            if b.used + size < b.size:
                return b.alloc(size, self.alignment)

        # No space left, alloc a new block
        return self._alloc_block(size).alloc(size, self.alignment)

    def advance(self) -> None:
        for b in self.blocks:
            b.advance()


_channels_dtype_int_bgra_to_format_table: Dict[Tuple[int, np.dtype, bool, bool], Format] = {
    # normalized formats
    (1, np.dtype(np.uint8), False, False): Format.R8_UNORM,
    (2, np.dtype(np.uint8), False, False): Format.R8G8_UNORM,
    (3, np.dtype(np.uint8), False, False): Format.R8G8B8_UNORM,
    (4, np.dtype(np.uint8), False, False): Format.R8G8B8A8_UNORM,
    (1, np.dtype(np.int8), False, False): Format.R8_SNORM,
    (2, np.dtype(np.int8), False, False): Format.R8G8_SNORM,
    (3, np.dtype(np.int8), False, False): Format.R8G8B8_SNORM,
    (4, np.dtype(np.int8), False, False): Format.R8G8B8A8_SNORM,
    (1, np.dtype(np.uint16), False, False): Format.R16_UNORM,
    (2, np.dtype(np.uint16), False, False): Format.R16G16_UNORM,
    (3, np.dtype(np.uint16), False, False): Format.R16G16B16_UNORM,
    (4, np.dtype(np.uint16), False, False): Format.R16G16B16A16_UNORM,
    (1, np.dtype(np.int16), False, False): Format.R16_SNORM,
    (2, np.dtype(np.int16), False, False): Format.R16G16_SNORM,
    (3, np.dtype(np.int16), False, False): Format.R16G16B16_SNORM,
    (4, np.dtype(np.int16), False, False): Format.R16G16B16A16_SNORM,
    # normalized formats BGR
    (3, np.dtype(np.uint8), False, True): Format.B8G8R8_UNORM,
    (4, np.dtype(np.uint8), False, True): Format.B8G8R8A8_UNORM,
    (3, np.dtype(np.int8), False, True): Format.B8G8R8_SNORM,
    (4, np.dtype(np.int8), False, True): Format.B8G8R8A8_SNORM,
    # integer formats
    (1, np.dtype(np.uint8), True, False): Format.R8_UINT,
    (2, np.dtype(np.uint8), True, False): Format.R8G8_UINT,
    (3, np.dtype(np.uint8), True, False): Format.R8G8B8_UINT,
    (4, np.dtype(np.uint8), True, False): Format.R8G8B8A8_UINT,
    (1, np.dtype(np.int8), True, False): Format.R8_SINT,
    (2, np.dtype(np.int8), True, False): Format.R8G8_SINT,
    (3, np.dtype(np.int8), True, False): Format.R8G8B8_SINT,
    (4, np.dtype(np.int8), True, False): Format.R8G8B8A8_SINT,
    (1, np.dtype(np.uint16), True, False): Format.R16_UINT,
    (2, np.dtype(np.uint16), True, False): Format.R16G16_UINT,
    (3, np.dtype(np.uint16), True, False): Format.R16G16B16_UINT,
    (4, np.dtype(np.uint16), True, False): Format.R16G16B16A16_UINT,
    (1, np.dtype(np.int16), True, False): Format.R16_SINT,
    (2, np.dtype(np.int16), True, False): Format.R16G16_SINT,
    (3, np.dtype(np.int16), True, False): Format.R16G16B16_SINT,
    (4, np.dtype(np.int16), True, False): Format.R16G16B16A16_SINT,
    (1, np.dtype(np.uint32), True, False): Format.R32_UINT,
    (2, np.dtype(np.uint32), True, False): Format.R32G32_UINT,
    (3, np.dtype(np.uint32), True, False): Format.R32G32B32_UINT,
    (4, np.dtype(np.uint32), True, False): Format.R32G32B32A32_UINT,
    (1, np.dtype(np.int32), True, False): Format.R32_SINT,
    (2, np.dtype(np.int32), True, False): Format.R32G32_SINT,
    (3, np.dtype(np.int32), True, False): Format.R32G32B32_SINT,
    (4, np.dtype(np.int32), True, False): Format.R32G32B32A32_SINT,
    # integer formats BGR
    (3, np.dtype(np.uint8), True, True): Format.B8G8R8_UINT,
    (4, np.dtype(np.uint8), True, True): Format.B8G8R8A8_UINT,
    (3, np.dtype(np.int8), True, True): Format.B8G8R8_SINT,
    (4, np.dtype(np.int8), True, True): Format.B8G8R8A8_SINT,
    # float formats
    (1, np.dtype(np.float16), False, False): Format.R16_SFLOAT,
    (2, np.dtype(np.float16), False, False): Format.R16G16_SFLOAT,
    (3, np.dtype(np.float16), False, False): Format.R16G16B16_SFLOAT,
    (4, np.dtype(np.float16), False, False): Format.R16G16B16A16_SFLOAT,
    (1, np.dtype(np.float32), False, False): Format.R32_SFLOAT,
    (2, np.dtype(np.float32), False, False): Format.R32G32_SFLOAT,
    (3, np.dtype(np.float32), False, False): Format.R32G32B32_SFLOAT,
    (4, np.dtype(np.float32), False, False): Format.R32G32B32A32_SFLOAT,
    (1, np.dtype(np.float64), False, False): Format.R64_SFLOAT,
    (2, np.dtype(np.float64), False, False): Format.R64G64_SFLOAT,
    (3, np.dtype(np.float64), False, False): Format.R64G64B64_SFLOAT,
    (4, np.dtype(np.float64), False, False): Format.R64G64B64A64_SFLOAT,
}

_format_to_channels_dtype_int_bgra_table: Dict[Format, Tuple[int, np.dtype, bool, bool]] = {
    v: k for k, v in _channels_dtype_int_bgra_to_format_table.items()
}

_format_to_srgb_format_table: Dict[Format, Format] = {
    Format.R8_UNORM: Format.R8_SRGB,
    Format.R8G8_UNORM: Format.R8G8_SRGB,
    Format.R8G8B8_UNORM: Format.R8G8B8_SRGB,
    Format.R8G8B8A8_UNORM: Format.R8G8B8A8_SRGB,
}


def format_from_channels_dtype(channels: int, dtype: np.dtype, integer: bool = False, bgra: bool = False) -> Format:
    try:
        return _channels_dtype_int_bgra_to_format_table[(channels, dtype, integer, bgra)]
    except KeyError:
        raise RuntimeError(
            f"Channels ({channels}) and dtype ({dtype}) combination does not match any format"
        ) from KeyError


def cull_mode_opposite_face(mode: CullMode) -> CullMode:
    if mode == CullMode.FRONT:
        return CullMode.BACK
    elif mode == CullMode.BACK:
        return CullMode.FRONT
    else:
        return mode


def readback_buffer(device: Device, buf: Buffer) -> NDArray[np.uint8]:
    host_buffer = Buffer(device, buf.size, BufferUsageFlags.TRANSFER_DST, AllocType.HOST)
    with device.sync_commands() as cmd:
        cmd.copy_buffer(buf, host_buffer)
        cmd.memory_barrier(MemoryUsage.TRANSFER_SRC, MemoryUsage.HOST_READ)
    array = np.frombuffer(host_buffer, np.uint8).copy()
    host_buffer.destroy()
    return array


def readback_image(device: Device, img: Image, new_layout: ImageLayout) -> NDArray[Any]:
    channels, dtype, _, _ = _format_to_channels_dtype_int_bgra_table[img.format]
    shape = (img.height, img.width, channels)

    buffer = Buffer(device, int(np.prod(shape)) * dtype.itemsize, BufferUsageFlags.TRANSFER_DST, AllocType.HOST)
    with device.sync_commands() as cmd:
        cmd.image_barrier(img, ImageLayout.TRANSFER_SRC_OPTIMAL, MemoryUsage.NONE, MemoryUsage.TRANSFER_SRC)
        cmd.copy_image_to_buffer(img, buffer)
        cmd.image_barrier(img, new_layout, MemoryUsage.TRANSFER_SRC, MemoryUsage.SHADER_READ_ONLY)
        cmd.memory_barrier(MemoryUsage.TRANSFER_SRC, MemoryUsage.HOST_READ)
    array = np.frombuffer(buffer, dtype).copy().reshape(shape)
    buffer.destroy()

    return array


def readback_image_mips(device: Device, img: Image, new_layout: ImageLayout) -> List[NDArray[Any]]:
    channels, dtype, _, _ = _format_to_channels_dtype_int_bgra_table[img.format]

    buffers: List[Buffer] = []
    for m in range(img.mip_levels):
        w, h = max(img.width >> m, 1), max(img.height >> m, 1)
        buffers.append(
            Buffer(device, w * h * channels * dtype.itemsize, BufferUsageFlags.TRANSFER_DST, AllocType.HOST)
        )

    with device.sync_commands() as cmd:
        cmd.image_barrier(img, ImageLayout.TRANSFER_SRC_OPTIMAL, MemoryUsage.NONE, MemoryUsage.TRANSFER_SRC)
        for m, buffer in zip(range(img.mip_levels), buffers):
            w, h = max(img.width >> m, 1), max(img.height >> m, 1)
            cmd.copy_image_to_buffer_range(img, buffer, w, h, image_mip=m)
        cmd.image_barrier(img, new_layout, MemoryUsage.TRANSFER_SRC, MemoryUsage.SHADER_READ_ONLY)
        cmd.memory_barrier(MemoryUsage.TRANSFER_SRC, MemoryUsage.HOST_READ)

    arrays: List[NDArray[Any]] = []
    for m, b in zip(range(img.mip_levels), buffers):
        w, h = max(img.width >> m, 1), max(img.height >> m, 1)
        shape = (h, w, channels)
        arrays.append(np.frombuffer(b, dtype).copy().reshape(shape))
        b.destroy()

    return arrays


def to_srgb_format(format: Format) -> Format:
    try:
        return _format_to_srgb_format_table[format]
    except KeyError:
        raise RuntimeError(f"{format} does not have a corresponding sRGB format") from KeyError


def get_min_max_and_required_subgroup_size(
    device: Device, stage: Stage, preferred_size: int, group_size: Optional[int] = None
) -> Tuple[int, int, Optional[int]]:
    if not is_pow_2(preferred_size):
        raise RuntimeError(f"Preferred size must be a power of 2. Got: {preferred_size}")

    control_props = device.device_properties.subgroup_size_control_properties
    if device.features & DeviceFeatures.SUBGROUP_SIZE_CONTROL:
        if device.subgroup_size_control and control_props.required_subgroup_size_stages & stage:
            min_subgroup_size = control_props.min_subgroup_size
            if group_size is not None:
                min_subgroup_size = group_size // control_props.max_compute_workgroup_subgroups
                if min_subgroup_size > control_props.max_subgroup_size:
                    raise RuntimeError(
                        f"Minimum subgroup size ({min_subgroup_size}) due to max compute workgroup subgroups ({control_props.max_compute_workgroup_subgroups}) for requested group size {group_size} is larger than maximum allowed subgroup size {control_props.max_subgroup_size}."
                    )
            subgroup_size = np.clip(preferred_size, min_subgroup_size, control_props.max_subgroup_size)
            return subgroup_size, subgroup_size, subgroup_size
        else:
            return control_props.min_subgroup_size, control_props.max_subgroup_size, None
    else:
        return (
            device.device_properties.subgroup_properties.subgroup_size,
            device.device_properties.subgroup_properties.subgroup_size,
            None,
        )



import os
import sys

import cv2
import numpy as np
from pyglm.glm import normalize, quatLookAtRH, vec3
from pyxpg import *

from ambra.config import CameraConfig, Config, GuiConfig, PlaybackConfig, RendererConfig
from ambra.geometry import create_axis3d_lines_and_colors, create_sphere
from ambra.lights import DirectionalLight, DirectionalShadowSettings, EnvironmentLight, UniformEnvironmentLight
from ambra.materials import PBRMaterial
from ambra.primitives3d import Lines, Mesh
from ambra.property import ArrayBufferProperty
from ambra.viewer import Viewer

os.environ["OPENCV_IO_ENABLE_OPENEXR"] = "1"


def main():
    viewer = Viewer(
        config=Config(
            playback=PlaybackConfig(
                playing=True,
            ),
            camera=CameraConfig(
                position=vec3(-7, 7, 1.7),
                target=vec3(0),
            ),
            world_up=(0, 0, 1),
            gui=GuiConfig(
                stats=True,
                playback=True,
            ),
            renderer=RendererConfig(msaa_samples=4, background_color=(0.1, 0.1, 0.1, 1.0)),
        ),
    )

    v, n, f = create_sphere(0.5, rings=32, sectors=64)
    p_v = ArrayBufferProperty(v[np.newaxis], np.float32)
    p_n = ArrayBufferProperty(n[np.newaxis], np.float32)
    p_f = ArrayBufferProperty(f[np.newaxis], np.uint32)

    spheres = []
    for y in range(8):
        for x in range(8):
            t_x = ((x + 0.5) / 8 * 2.0 - 1.0) * 5
            t_y = ((y + 0.5) / 8 * 2.0 - 1.0) * 5
            if (y ^ x) & 1:
                m = PBRMaterial((1.0, 1.0, 1.0), (x + 0.5) / 8, (y + 0.5) / 8)
            else:
                m = PBRMaterial((1.0, 0.0, 0.0), (x + 0.5) / 8, (y + 0.5) / 8)
            sphere = Mesh(p_v, p_f, normals=p_n, translation=(t_x, t_y, 1), material=m)
            spheres.append(sphere)

    o = Lines(*create_axis3d_lines_and_colors(), 4.0)

    light_position = vec3(5, 6, 7) * 4.0
    light_target = vec3(0, 0, 0)

    signs = [
        vec3(-1, -1, 1),
        vec3(1, -1, 1),
        vec3(-1, 1, 1),
        vec3(1, 1, 1),
    ]

    lights = []
    for s in signs:
        rotation = quatLookAtRH(normalize(light_target - light_position * s), vec3(0, 1, 0))
        lights.append(
            DirectionalLight(
                np.array([0.3, 0.3, 0.3]),
                shadow_settings=DirectionalShadowSettings(casts_shadow=False),
                translation=light_position * s,
                rotation=rotation,
            )
        )

    equirectangular = cv2.imread(sys.argv[1], cv2.IMREAD_ANYCOLOR | cv2.IMREAD_ANYDEPTH | cv2.IMREAD_COLOR_RGB)
    lights.append(EnvironmentLight.from_equirectangular(equirectangular))
    # lights.append(UniformEnvironmentLight((0.1, 0.1, 0.1)))

    viewer.scene.objects.extend(spheres + [o] + lights)
    viewer.run()


if __name__ == "__main__":
    main()



from pathlib import Path

from pyxpg import imgui

from ambra.config import Config, DefaultFontPreference, GuiConfig
from ambra.viewer import Viewer


class CustomViewer(Viewer):
    def on_gui(self):
        imgui.push_font(font, 15)
        super().on_gui()
        imgui.pop_font()


viewer = CustomViewer(
    "Fonts",
    config=Config(
        gui=GuiConfig(
            stats=True,
            default_font_preference=DefaultFontPreference.VECTOR,
        ),
    ),
)

path = Path(__file__).absolute().parent.parent.parent.joinpath("Roboto-Medium.ttf")
font = viewer.gui.add_font_ttf("Roboto", path.read_bytes())

viewer.run()



import sys
from pathlib import Path

import numpy as np
from pyglm.glm import vec3
from pyxpg import AllocType, Buffer, BufferUsageFlags, Format, Image, ImageLayout, ImageUsageFlags, MemoryUsage, imgui

from ambra.config import Config, GuiConfig
from ambra.lights import DirectionalLight, DirectionalShadowSettings
from ambra.marching_cubes import MarchingCubesPipeline
from ambra.primitives3d import Grid, GridType, MarchingCubesMesh, Mesh
from ambra.utils.gpu import readback_buffer, view_bytes
from ambra.utils.hook import hook
from ambra.viewer import Viewer

if False:
    x, y, z = 128, 128, 128
    sdf = np.zeros((x, y, z), np.float32)
    size = (1.0, 1.0, 1.0)
else:
    vol_path = Path(sys.argv[1])
    sdf: np.ndarray = np.load(vol_path)["volume"]
    size = np.array(sdf.shape[::-1], np.float32) / max(sdf.shape)
    z, y, x = sdf.shape

    if True:
        # (A, Z, Y, X, C)
        sdf = sdf.reshape((1, z, y, x, 1))

        N = 32
        sdf = np.repeat(sdf, N, axis=0)
        for i in range(N):
            sdf[i] += i / N

    if False:
        # Dragon big
        # 1662892 positions
        # 7638876 primitives
        #
        # 1654432 area
        #
        # expected_positions = max(area * 2, 1024)
        # expected_triangles = max(area * 10, 5 * 1024)
        #
        # Dragon small
        #  65248
        # 300216
        #
        # 102634 area
        area = (x * y + x * z + y * z) * 2
        print(x, y, z, area)


class CustomViewer(Viewer):
    @hook
    def on_gui(self):
        global m
        if isinstance(m, MarchingCubesMesh):
            if imgui.begin("MC")[0]:
                u, m.level = imgui.slider_float("Level", m.level, -5, 5)
                if u:
                    m.invalidate_surface()
            imgui.end()


v = CustomViewer(config=Config(gui=GuiConfig(stats=True)))

print(v.device.features)
print(v.device.subgroup_size_control)
print(v.device.compute_full_subgroups)
print(v.device.device_properties.subgroup_size_control_properties.min_subgroup_size)
print(v.device.device_properties.subgroup_size_control_properties.max_subgroup_size)
print(v.device.device_properties.subgroup_size_control_properties.max_compute_workgroup_subgroups)
print(v.device.device_properties.subgroup_size_control_properties.required_subgroup_size_stages)

if False:
    # print(v.renderer.device.device_properties.subgroup_properties.subgroup_size)
    pipeline = MarchingCubesPipeline(v.renderer)

    sdf_buf = Buffer.from_data(
        v.device, view_bytes(sdf), BufferUsageFlags.TRANSFER_SRC, AllocType.HOST, name="sdf-buf"
    )
    sdf_gpu = Image(
        v.device,
        x,
        y,
        depth=z,
        format=Format.R32_SFLOAT,
        usage_flags=ImageUsageFlags.STORAGE | ImageUsageFlags.TRANSFER_DST,
        alloc_type=AllocType.DEVICE,
        name="sdf",
    )

    with v.device.sync_commands() as cmd:
        cmd.image_barrier(sdf_gpu, ImageLayout.TRANSFER_DST_OPTIMAL, MemoryUsage.NONE, MemoryUsage.TRANSFER_DST)
        cmd.copy_buffer_to_image(sdf_buf, sdf_gpu)
        cmd.image_barrier(sdf_gpu, ImageLayout.GENERAL, MemoryUsage.TRANSFER_DST, MemoryUsage.COMPUTE_SHADER)

    res = pipeline.run_sync(v.renderer, sdf_gpu, size, 0.0)

    positions = readback_buffer(v.device, res.positions).view(np.float32).reshape((-1, 3))
    normals = readback_buffer(v.device, res.normals).view(np.float32).reshape((-1, 3))
    indices = readback_buffer(v.device, res.indices).view(np.uint32)

    print(positions.shape)
    print(normals.shape)
    print(indices.shape)
    m = Mesh(positions, indices, normals)
else:
    m = MarchingCubesMesh(sdf[:, :, :], size)

l1 = DirectionalLight.look_at(
    vec3(5, 6, 7) * 4.0,
    vec3(0, 0, 0),
    vec3(0, 0, 1),
    np.array([1.0, 1.0, 1.0]),
    shadow_settings=DirectionalShadowSettings(half_extent=10.0, z_near=1.0, z_far=100),
)
l2 = DirectionalLight.look_at(
    vec3(-5, -6, 7) * 4.0,
    vec3(0, 0, 0),
    vec3(0, 0, 1),
    np.array([1.0, 1.0, 1.0]),
    shadow_settings=DirectionalShadowSettings(half_extent=10.0, z_near=1.0, z_far=100),
)

g = Grid.transparent_black_lines((100, 100), GridType.XY_PLANE)

v.scene.objects.extend([m, l1, l2, g])
v.run()



import sys
from dataclasses import dataclass
from pathlib import Path

import numpy as np
from pyxpg import imgui
from scipy.spatial.transform import Rotation

from ambra.config import CameraConfig, Config, GuiConfig, RendererConfig
from ambra.geometry import create_axis3d_lines_and_colors
from ambra.primitives3d import GaussianSplats, GaussianSplatsRenderFlags, Lines
from ambra.property import ArrayBufferProperty, UploadSettings
from ambra.utils.hook import hook
from ambra.viewer import Viewer


@dataclass
class Splats:
    position: np.ndarray
    sh: np.ndarray
    opacity: np.ndarray
    scale: np.ndarray
    rotation: np.ndarray


def load_ply(path: Path, sh_degree: int = 1):
    with open(path, "rb") as f:
        # Read header.
        head = f.readline().decode("utf-8").strip().lower()
        if head != "ply":
            raise ValueError(f"Not a ply file: {head}")

        encoding = f.readline().decode("utf-8").strip().lower()
        if "binary_little_endian" not in encoding:
            raise ValueError(f"Invalid encoding: {encoding}")

        elements = f.readline().decode("utf-8").strip().lower()
        count = int(elements.split()[2])

        # Read until end of header.
        # TODO: properly parse header
        while f.readline().decode("utf-8").strip().lower() != "end_header":
            pass

        # Number of 32 bit floats used to encode Spherical Harmonics coefficients.
        # The last multiplication by 3 is because we have 3 components (RGB) for each coefficient.
        sh_coeffs = (sh_degree + 1) * (sh_degree + 1) * 3

        # Position (vec3), normal (vec3), spherical harmonics (sh_coeffs), opacity (float),
        # scale (vec3) and rotation (quaternion). All values are float32 (4 bytes).
        size = count * (3 + 3 + sh_coeffs + 1 + 3 + 4) * 4

        data = f.read(size)
        arr = np.frombuffer(data, dtype=np.float32).reshape((count, -1))

        # Positions.
        position = arr[:, :3].copy()

        # Currently we don't need normals for rendering.
        # normal = arr[:, 3:6].copy()

        # Spherical harmonic coefficients.
        sh = arr[:, 6 : 6 + sh_coeffs].copy()

        # Activate alpha: sigmoid(alpha).
        opacity = 1.0 / (1.0 + np.exp(-arr[:, 6 + sh_coeffs]))

        # Exponentiate scale.
        scale = np.exp(arr[:, 7 + sh_coeffs : 10 + sh_coeffs])

        # Normalize quaternions.
        rotation = arr[:, 10 + sh_coeffs : 14 + sh_coeffs].copy()
        rotation /= np.linalg.norm(rotation, ord=2, axis=1)[..., np.newaxis]

        # Convert from wxyz to xyzw.
        rotation = np.roll(rotation, -1, axis=1)

        return Splats(position, sh, opacity, scale, rotation)


splats = load_ply(sys.argv[1], 3)

if False:
    positions = np.array([[1, 0, 0]], np.float32)
    colors = np.array([[0.5, 0, 0.0, 1]], np.float32)
    sh = np.zeros((1, 13), np.uint8)
    covariances = np.array([[1, 0, 0, 2, 0, 3]], np.float32) * 1
else:
    positions = splats.position
    # positions = np.vstack((np.linspace(positions * 0.5, positions, 30), np.linspace(positions, positions * 0.5, 30)))

    SH_C0 = 0.28209479177387814
    colors = np.hstack((splats.sh[:, :3] * SH_C0 + 0.5, splats.opacity[..., np.newaxis]))
    sh = np.transpose(
        np.clip(np.rint((splats.sh[:, 3:] * 0.5 + 0.5) * 255.0), 0.0, 255.0).astype(np.uint8).reshape((-1, 3, 15)),
        # splats.sh[:, 3:].astype(np.float16).reshape((-1, 3, 15)),
        # splats.sh[:, 3:].reshape((-1, 3, 15)),
        (0, 2, 1),
    ).reshape((-1, 45))

    rots = Rotation.from_quat(splats.rotation).as_matrix()
    scale = np.zeros((splats.scale.shape[0], 3, 3), np.float32)
    scale[:, 0, 0] = splats.scale[:, 0]
    scale[:, 1, 1] = splats.scale[:, 1]
    scale[:, 2, 2] = splats.scale[:, 2]

    covariance_matrices: np.ndarray = np.matmul(rots, scale)
    transformed_covariance: np.ndarray = np.matmul(covariance_matrices, np.transpose(covariance_matrices, (0, 2, 1)))
    covariances = transformed_covariance.reshape((-1, 9))[:, (0, 1, 2, 4, 5, 8)]

# positions = ArrayBufferProperty(
#     positions,
#     np.float32,
#     upload=UploadSettings(preupload=False, cpu_prefetch_count=2, gpu_prefetch_count=2, async_load=True),
# )
gs = GaussianSplats(positions, colors, sh, covariances)


class CustomViewer(Viewer):
    @hook
    def on_gui(self):
        if imgui.begin("Gaussian splatting")[0]:
            _, v = imgui.checkbox("Disable opacity", (gs.flags & GaussianSplatsRenderFlags.DISABLE_OPACITY) != 0)
            if v:
                gs.flags |= GaussianSplatsRenderFlags.DISABLE_OPACITY
            else:
                gs.flags &= ~GaussianSplatsRenderFlags.DISABLE_OPACITY

            _, v = imgui.checkbox(
                "Show SH only", (gs.flags & GaussianSplatsRenderFlags.SHOW_SPHERICAL_HARMONICS_ONLY) != 0
            )
            if v:
                gs.flags |= GaussianSplatsRenderFlags.SHOW_SPHERICAL_HARMONICS_ONLY
            else:
                gs.flags &= ~GaussianSplatsRenderFlags.SHOW_SPHERICAL_HARMONICS_ONLY

            _, v = imgui.checkbox("Point cloud mode", (gs.flags & GaussianSplatsRenderFlags.POINT_CLOUD_MODE) != 0)
            if v:
                gs.flags |= GaussianSplatsRenderFlags.POINT_CLOUD_MODE
            else:
                gs.flags &= ~GaussianSplatsRenderFlags.POINT_CLOUD_MODE
            _, gs.alpha_cull_threshold = imgui.drag_float(
                "Alpha cull threshold", gs.alpha_cull_threshold, v_speed=0.01, v_min=0.0, v_max=1.0
            )
            _, gs.frustum_dilation = imgui.drag_float(
                "Frustum dilation", gs.frustum_dilation, v_speed=0.01, v_min=0.0, v_max=1.0
            )
            _, gs.splat_scale = imgui.drag_float("Splat scale", gs.splat_scale, v_speed=0.01, v_min=0.0, v_max=10.0)
        imgui.end()


v = CustomViewer(
    config=Config(
        gui=GuiConfig(
            stats=True,
            inspector=True,
            renderer=True,
        ),
        world_up=(0, -1, 0),
        renderer=RendererConfig(
            background_color=(0, 0, 0, 1),
        ),
        camera=CameraConfig(position=(-1, -1, -1)),
    )
)
v.scene.objects.append(gs)

line = Lines(*create_axis3d_lines_and_colors(), 1)

v.scene.objects.extend([line])

v.run()



# Copyright Dario Mylonopoulos
# SPDX-License-Identifier: MIT

from dataclasses import dataclass

from pyglm.glm import (
    decompose,
    inverse,
    mat4,
    mat4_cast,
    normalize,
    quat,
    quat_cast,
    quatLookAtLH,
    quatLookAtRH,
    vec3,
    vec4,
)

from .config import Handedness


@dataclass
class RigidTransform3D:
    translation: vec3
    rotation: quat

    @classmethod
    def look_at(cls, position: vec3, target: vec3, up: vec3, handedness: Handedness) -> "RigidTransform3D":
        d = normalize(target - position)
        if handedness == Handedness.RIGHT_HANDED:
            inv_rot = quatLookAtRH(d, up)
        else:
            inv_rot = quatLookAtLH(d, up)
        rot = inverse(inv_rot)

        return cls(
            translation=rot * -position,  # type: ignore
            rotation=rot,  # type: ignore
        )

    @classmethod
    def from_mat4(cls, m: mat4) -> "RigidTransform3D":
        return cls(
            rotation=quat_cast(m),
            translation=vec3(m[3, 0], m[3, 1], m[3, 2]),
        )

    def as_mat4(self) -> mat4:
        m = mat4_cast(self.rotation)
        m[3, 0] = self.translation.x
        m[3, 1] = self.translation.y
        m[3, 2] = self.translation.z
        return m

    @classmethod
    def identity(cls) -> "RigidTransform3D":
        return cls(
            translation=vec3(0),
            rotation=quat(1, 0, 0, 0),
        )

    def inverse(self) -> "RigidTransform3D":
        inverse_rotation = inverse(self.rotation)
        return RigidTransform3D(translation=-(inverse_rotation * self.translation), rotation=inverse_rotation)  # type:ignore

    def __matmul__(self, other: "RigidTransform3D") -> "RigidTransform3D":
        return RigidTransform3D(
            rotation=self.rotation * other.rotation,
            translation=self.rotation * other.translation + self.translation,
        )

    def copy(self) -> "RigidTransform3D":
        return RigidTransform3D(translation=vec3(self.translation), rotation=quat(self.rotation))


@dataclass
class Transform3D:
    translation: vec3
    rotation: quat
    scale: vec3

    @classmethod
    def identity(cls) -> "Transform3D":
        return cls(
            translation=vec3(0),
            rotation=quat(1, 0, 0, 0),
            scale=vec3(1),
        )

    @classmethod
    def from_mat4(cls, m: mat4) -> "Transform3D":
        t = Transform3D(vec3(), quat(), vec3())
        skew = vec3()
        perspective = vec4()
        decompose(mat4(m), t.scale, t.rotation, t.translation, skew, perspective)
        return t

    def as_mat4(self) -> mat4:
        m = mat4_cast(self.rotation)
        m[0] *= self.scale[0]
        m[1] *= self.scale[1]
        m[2] *= self.scale[2]
        m[3, 0] = self.translation.x
        m[3, 1] = self.translation.y
        m[3, 2] = self.translation.z
        return m

    def copy(self) -> "Transform3D":
        return Transform3D(translation=vec3(self.translation), rotation=quat(self.rotation), scale=vec3(self.scale))



# Copyright Dario Mylonopoulos
# SPDX-License-Identifier: MIT

import atexit
from queue import Queue
from threading import Event, Thread
from typing import Any, Callable, Dict, Generic, Optional, Tuple, TypeVar

O = TypeVar("O")


class PromiseError(RuntimeError):
    pass


class Promise(Generic[O]):
    def __init__(self) -> None:
        self.obj: Optional[O] = None
        self.exception: Optional[Exception] = None
        self.event: Event = Event()

    def is_set(self) -> bool:
        return self.event.is_set()

    def clear(self) -> None:
        self.event.clear()

    def set(self, obj: O) -> None:
        self.obj = obj
        self.event.set()

    def set_exception(self, e: Exception) -> None:
        self.exception = e
        self.event.set()

    def get(self) -> O:
        self.event.wait()
        if self.exception is not None:
            raise PromiseError from self.exception
        return self.obj  # type: ignore


class ThreadPool:
    def __init__(self, num_workers: int):
        self.queue: Queue[
            Optional[
                Tuple[
                    Promise[Any],
                    Callable[[Tuple[Any, ...], Dict[str, Any]], Any],
                    Tuple[Any, ...],
                    Dict[str, Any],
                ]
            ]
        ] = Queue()
        self.workers = [
            Thread(None, self.__entry, f"ThreadPool.worker{i}", (i,), daemon=True) for i in range(num_workers)
        ]
        for w in self.workers:
            w.start()
        atexit.register(self.stop)

    def __entry(self, thread_index: int) -> None:
        while True:
            elem = self.queue.get()
            if elem is None:
                break

            promise, func, args, kwargs = elem
            try:
                # HACK: having this special thread_index variable here is a bit messy and mypy complains about it, maybe we can do something cleaner
                obj = func(*args, **kwargs, thread_index=thread_index)  # type: ignore
                promise.set(obj)
            except Exception as e:
                promise.set_exception(e)

    def submit(
        self,
        promise: Promise[Any],
        func: Callable[[Tuple[Any, ...], Dict[str, Any]], Any],
        *args: Any,
        **kwargs: Any,
    ) -> None:
        promise.event.clear()
        self.queue.put((promise, func, args, kwargs))

    def stop(self) -> None:
        for _ in range(len(self.workers)):
            self.queue.put(None)
        for w in self.workers:
            w.join()
        self.workers = []



from ambra.config import CameraConfig, Config, GuiConfig, PlaybackConfig
from ambra.geometry import create_axis3d_lines_and_colors
from ambra.keybindings import KeyMap, Modifiers, MouseButton, MouseButtonBinding
from ambra.primitives3d import Lines
from ambra.viewer import Viewer

viewer = Viewer(
    config=Config(
        wait_events=True,
        playback=PlaybackConfig(
            playing=True,
        ),
        gui=GuiConfig(
            stats=True,
        ),
        world_up=(0, 1, 0),
        camera=CameraConfig(
            position=(3, 3, 3),
            target=(0, 0, 0),
        ),
    ),
    key_map=KeyMap(
        camera_rotate=MouseButtonBinding(MouseButton.LEFT, Modifiers.NONE),
        camera_pan=MouseButtonBinding(MouseButton.RIGHT, Modifiers.NONE),
    ),
)

line = Lines(*create_axis3d_lines_and_colors(), 4)
viewer.scene.objects.extend([line])
viewer.run()



from dataclasses import dataclass
from pathlib import Path
from typing import Dict, List, Optional

import numpy as np
import pygltflib
from numpy.typing import NDArray
from PIL import Image
from pyglm.glm import mat4, quat, vec3

from ambra.transform3d import Transform3D


@dataclass
class Joint:
    name: str
    transform: Transform3D
    parent_index: int
    inverse_bind_matrix: mat4
    val: float = 1.0


@dataclass
class MeshData:
    # Geometry
    indices: Optional[NDArray] = None
    positions: Optional[NDArray] = None
    normals: Optional[NDArray] = None
    tangents: Optional[NDArray] = None
    uvs: Optional[NDArray] = None
    joint_indices: Optional[NDArray] = None
    weights: Optional[NDArray] = None

    # Material
    base_color_texture: Optional[NDArray] = None
    metallic_roughness_texture: Optional[NDArray] = None
    normal_texture: Optional[NDArray] = None
    ao_texture: Optional[NDArray] = None

    # Joints
    joints: List[Joint] = None


@dataclass
class Scene:
    meshes: List[MeshData]


def load(path: Path) -> Scene:
    gltf = pygltflib.GLTF2().load(path)

    meshes: List[MeshData] = []

    buffers = []
    for b in gltf.buffers:
        buf = path.parent.joinpath(b.uri).read_bytes()
        assert b.byteLength == len(buf), f"Expected: {b.byteLength}, Got: {len(buf)}"
        buffers.append(buf)

    def load_attribute(ai: int) -> NDArray:
        a = gltf.accessors[ai]
        assert a.bufferView is not None
        view = gltf.bufferViews[a.bufferView]

        data = memoryview(buffers[view.buffer])[view.byteOffset :][: view.byteLength][a.byteOffset :]

        dtype = {
            pygltflib.FLOAT: np.float32,
            pygltflib.UNSIGNED_SHORT: np.int16,
            pygltflib.BYTE: np.int8,
            pygltflib.UNSIGNED_BYTE: np.uint8,
            pygltflib.SHORT: np.int16,
            pygltflib.UNSIGNED_SHORT: np.uint16,
            pygltflib.UNSIGNED_INT: np.uint32,
        }[a.componentType]

        shape = {
            "SCALAR": (1,),
            "VEC2": (2,),
            "VEC3": (3,),
            "VEC4": (4,),
            "MAT4": (4, 4),
        }[a.type]

        if not data.c_contiguous:
            data = bytes(data)

        data_bytestride = np.prod(shape) * np.dtype(dtype).itemsize
        if view.byteStride is not None:
            assert view.byteStride == data_bytestride, f"{view.byteStride} -> {data_bytestride}"

        arr = np.frombuffer(data, dtype)
        if a.type != "SCALAR":
            arr = arr.reshape((-1, *shape))
        return arr[: a.count]

    def load_rec(ni: int, depth: int):
        node = gltf.nodes[ni]
        print("    " * depth + f"{ni}: {node}")

        joints: List[Joint] = []
        if node.skin is not None:
            skin = gltf.skins[node.skin]
            print("    " * (depth + 1) + ">" + str(skin))

            # Assume there is a single root joint and it's skeleton, this is not necessarily true but simiplifies
            # parsing.
            assert skin.skeleton is not None

            node_id_to_joint_index: Dict[int, int] = {}

            def load_joint_rec(ni: int, parent_index: int):
                node = gltf.nodes[ni]

                t = Transform3D(
                    translation=vec3(node.translation) if node.translation is not None else vec3(0.0),
                    rotation=quat(node.rotation[3], *node.rotation[:3])
                    if node.rotation is not None
                    else quat(1, 0, 0, 0),
                    # rotation=inverse(quat(node.rotation[3], *node.rotation[:3])) if node.rotation is not None else quat(1, 0, 0, 0),
                    # rotation=quat(node.rotation) if node.rotation is not None else quat(1, 0, 0, 0),
                    scale=vec3(node.scale) if node.scale is not None else vec3(1.0),
                )
                # print(t.rotation, node.rotation)
                # exit(1)
                j = Joint(node.name, t, parent_index, None)
                print(node.name, t)
                joint_index = len(joints)
                node_id_to_joint_index[ni] = joint_index
                joints.append(j)

                for c in node.children:
                    load_joint_rec(c, joint_index)

            load_joint_rec(skin.skeleton, -1)

            inverse_bind_matrices = load_attribute(skin.inverseBindMatrices)
            for i, j in enumerate(skin.joints):
                # Assume joints are given in traversal order, we would actually need to reorder here
                assert i == node_id_to_joint_index[j]

                joints[node_id_to_joint_index[j]].inverse_bind_matrix = mat4(inverse_bind_matrices[i])

        if node.mesh is not None:
            # Load mesh
            mesh = gltf.meshes[node.mesh]
            print("    " * (depth + 1) + ">" + str(mesh))

            for p in mesh.primitives:
                m = MeshData(joints=joints)

                print("    " * (depth + 2) + ">" + str(p))

                if p.indices is not None:
                    indices = gltf.accessors[p.indices]
                    print("    " * (depth + 3) + "|-> INDICES", p.indices, indices)

                    m.indices = load_attribute(p.indices)

                for k, v in p.attributes.__dict__.items():
                    if k.startswith("__"):
                        continue
                    if v is None:
                        continue
                    print("    " * (depth + 3) + "|-> " + k, v, gltf.accessors[v], flush=True)

                    if k == "POSITION":
                        m.positions = load_attribute(v)
                    elif k == "NORMAL":
                        m.normals = load_attribute(v)
                    elif k == "TANGENT":
                        m.tangents = load_attribute(v)
                    elif k == "TEXCOORD_0":
                        m.uvs = load_attribute(v)
                    elif k == "JOINTS_0":
                        m.joint_indices_0 = load_attribute(v)
                    elif k == "WEIGHTS_0":
                        m.weights_0 = load_attribute(v)
                    elif k == "JOINTS_1":
                        m.joint_indices_1 = load_attribute(v)
                    elif k == "WEIGHTS_1":
                        m.weights_1 = load_attribute(v)

                if p.material is not None:
                    material = gltf.materials[p.material]
                    print("    " * (depth + 3) + ">" + str(material))

                    if material.pbrMetallicRoughness is not None:

                        def load_image(texture_attribute):
                            img = None
                            if texture_attribute:
                                texture = gltf.textures[texture_attribute.index]
                                if texture.source is not None:
                                    image = gltf.images[texture.source]
                                    print("    " * (depth + 5) + ">" + str(image))
                                    img = np.array(Image.open(path.parent.joinpath(image.uri)))
                                if texture.sampler is not None:
                                    sampler = gltf.samplers[texture.sampler]
                                    print("    " * (depth + 5) + ">" + str(sampler))
                            return img

                        m.base_color_texture = load_image(material.pbrMetallicRoughness.baseColorTexture)
                        m.metallic_roughness_texture = load_image(
                            material.pbrMetallicRoughness.metallicRoughnessTexture
                        )
                    m.ao_texture = load_image(material.occlusionTexture)
                    m.normal_texture = load_image(material.normalTexture)

                meshes.append(m)

        for ci in node.children:
            load_rec(ci, depth + 1)

    for s in gltf.scenes:
        for ni in s.nodes:
            load_rec(ni, 0)

    return Scene(meshes)



import numpy as np

from ambra.config import CameraConfig, Config, PlaybackConfig
from ambra.primitives3d import Image
from ambra.property import AnimationBoundary, ArrayBufferProperty
from ambra.utils.gpu import Format
from ambra.viewer import Viewer

viewer = Viewer(
    config=Config(
        playback=PlaybackConfig(
            playing=True,
        ),
        camera=CameraConfig(
            position=(0.5, 0.5, -2),
            target=(0, 0, 0),
        ),
        world_up=(0, 1, 0),
    ),
)

translation = ArrayBufferProperty(
    np.linspace(np.array([0, 0, 0]), np.array([5, 0, 0]), 100),
    np.float32,
)

H, W = 32, 64
start = np.zeros((H, W, 4))
end = np.zeros((H, W, 4))
for i in range(H):
    for j in range(W):
        start[i, j, 0] = (i + 0.5) / H
        start[i, j, 1] = (j + 0.5) / W
        start[i, j, 3] = 0

        end[i, j, 0] = (j + 0.5) / W
        end[i, j, 2] = (i + 0.5) / H
        end[i, j, 3] = 0
image = (np.linspace(start, end, 100) * 255).astype(np.uint8)

img = Image(image, translation=translation, scale=(1, H / W, 1))
viewer.scene.objects.extend([img])

viewer.run()



import numpy as np
from pyxpg import AllocType, Buffer, BufferUsageFlags, MemoryUsage, PipelineStageFlags, QueryPool, QueryType

from ambra.config import Config
from ambra.gpu_sorting import GpuSortingPipeline, SortDataType, SortOptions
from ambra.viewer import Viewer

v = Viewer(
    config=Config(
        window=False,
    )
)

# Prepare pipleine
sort = GpuSortingPipeline(
    v.renderer,
    SortOptions(
        key_type=SortDataType.UINT32,
        payload_type=SortDataType.UINT32,
        unsafe_has_forward_thread_progress_guarantee=False,
    ),
)

# Create random keys
np.random.seed(42)
N = 1 << 27
keys_array = np.random.randint(0, 0xFFFFFFFF, N, np.uint32)

# Alocate buffers
keys = Buffer.from_data(
    v.device,
    keys_array,
    BufferUsageFlags.STORAGE | BufferUsageFlags.TRANSFER_DST | BufferUsageFlags.TRANSFER_SRC,
    AllocType.DEVICE_MAPPED_WITH_FALLBACK,
    name="keys",
)
keys_alt = Buffer(v.device, keys.size, BufferUsageFlags.STORAGE, AllocType.DEVICE, name="keys-alt")

payload = Buffer.from_data(
    v.device,
    np.arange(N, dtype=np.uint32),
    BufferUsageFlags.STORAGE | BufferUsageFlags.TRANSFER_DST | BufferUsageFlags.TRANSFER_SRC,
    AllocType.DEVICE_MAPPED_WITH_FALLBACK,
    name="payload",
)
payload_alt = Buffer(v.device, payload.size, BufferUsageFlags.STORAGE, AllocType.DEVICE, name="payload-alt")

BENCHMARK = True
if BENCHMARK:
    for i in range(27):
        n = 1 << i
        dt = 1000_000_000
        for _ in range(5):
            pool = QueryPool(v.device, QueryType.TIMESTAMP, 2)
            v.device.reset_query_pool(pool)
            with v.device.sync_commands() as cmd:
                cmd.write_timestamp(pool, 0, PipelineStageFlags.TOP_OF_PIPE)
                sort.run(v.renderer, cmd, n, keys, keys_alt, payload, payload_alt)
                cmd.memory_barrier(MemoryUsage.COMPUTE_SHADER, MemoryUsage.TRANSFER_SRC)
                cmd.write_timestamp(pool, 1, PipelineStageFlags.BOTTOM_OF_PIPE)
            res = pool.wait_results(0, 2)
            dt = min(dt, res[1] - res[0])
        dt_ns = dt * v.device.timestamp_period_ns
        print(f"{n:12}: {dt_ns * 1e-3:12.03f}us  {n / (dt_ns)} B/s")
else:
    readback_keys = Buffer(
        v.device,
        keys.size,
        BufferUsageFlags.STORAGE | BufferUsageFlags.TRANSFER_DST,
        AllocType.HOST,
        name="readback-keys",
    )
    readback_payload = Buffer(
        v.device,
        payload.size,
        BufferUsageFlags.STORAGE | BufferUsageFlags.TRANSFER_DST,
        AllocType.HOST,
        name="readback-payload",
    )

    # Sort and readback
    with v.device.sync_commands() as cmd:
        sort.run(v.renderer, cmd, N, keys, keys_alt, payload, payload_alt)
        cmd.memory_barrier(MemoryUsage.COMPUTE_SHADER, MemoryUsage.TRANSFER_SRC)
        cmd.copy_buffer(keys, readback_keys)
        cmd.copy_buffer(payload, readback_payload)
        cmd.memory_barrier(MemoryUsage.TRANSFER_SRC, MemoryUsage.HOST_READ)

    # Check results
    readback_keys_array = np.frombuffer(readback_keys.data, np.uint32)
    readback_payload_array = np.frombuffer(readback_payload.data, np.uint32)

    indices_array = np.argsort(keys_array, kind="stable").astype(np.uint32)
    sorted_keys = keys_array[indices_array]

    assert np.array_equal(readback_keys_array, sorted_keys)
    assert np.array_equal(readback_payload_array, indices_array)



import numpy as np

from ambra.config import (
    CameraConfig,
    CameraProjection,
    Config,
    GuiConfig,
    PlaybackConfig,
    RendererConfig,
    UploadMethod,
)
from ambra.primitives2d import Image
from ambra.property import ImageProperty, UploadSettings
from ambra.utils.gpu import Format
from ambra.viewer import Viewer

viewer = Viewer(
    config=Config(
        playback=PlaybackConfig(
            playing=True,
        ),
        renderer=RendererConfig(
            background_color=(0, 0, 0, 1),
            # force_image_upload_method=UploadMethod.GRAPHICS_QUEUE,
            # force_image_upload_method=UploadMethod.TRANSFER_QUEUE,
        ),
        gui=GuiConfig(
            stats=True,
            playback=True,
            inspector=True,
            renderer=True,
        ),
        camera=CameraConfig(
            projection=CameraProjection.ORTHOGRAPHIC,
            ortho_half_extents=(2, 2),
            position=(0, 0, 0),
            target=(0, 0, 1),
        ),
        world_up=(0, 1, 0),
    ),
)

N = 100
W = 512
H = 256
C = 4


class GeneratedStreamingProperty(ImageProperty):
    def get_frame_by_index(self, frame_index: int, thread_index: int = -1):
        arr = np.full((H, W, C), frame_index, np.uint8)
        return arr


image_gen = GeneratedStreamingProperty(
    W,
    H,
    Format.R8G8B8A8_UNORM,
    N,
    upload=UploadSettings(
        preupload=False,
        async_load=True,
        cpu_prefetch_count=2,
        gpu_prefetch_count=2,
    ),
)

image = Image(image_gen)
viewer.scene.objects.append(image)

viewer.run()



import socket
import struct

s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
s.connect(("localhost", 9168))

name = "CLIENT".encode(encoding="utf-8")
name_length = struct.pack("<I", len(name))
s.send(b"AMBR" + name_length + name)

for i in range(10):
    data = f"Data is {i}".encode("utf-8") + bytearray(1024 * 1024)
    msg_header = struct.pack("<IIQ", 1 << 20, 0, len(data))
    s.send(msg_header + data)
s.send(struct.pack("<I", 0))
